# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import logging
import copy
import json
import math

import boto3

from .config import Config, validate_resource
from . import resources, templates, utils, cfn_yaml_tags
from .. import api_utils

"""
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-macros.html
Input:
{
    "region" : "us-east-1",
    "accountId" : "$ACCOUNT_ID",
    "fragment" : { ... },
    "transformId" : "$TRANSFORM_ID",
    "params" : { ... },
    "requestId" : "$REQUEST_ID",
    "templateParameterValues" : { ... }
}

Output:
{
    "requestId" : "$REQUEST_ID",
    "status" : "$STATUS",
    "fragment" : { ... }
}
"""
TRANSFORM_NAME = "AWS-SSO-Util-2020-11-08"
RESOURCE_TYPE = "SSOUtil::SSO::AssignmentGroup"

LOGGER = logging.getLogger(__name__)

SESSION = boto3.Session()
IDS = api_utils.Ids(SESSION)

def is_macro_template(template):
    if "Transform" not in template:
        return False
    transform = template["Transform"]
    if transform == TRANSFORM_NAME:
        return True
    if isinstance(transform, list) and TRANSFORM_NAME in transform:
        return True
    return False

def process_template(template,
        session,
        ids: api_utils.Ids,
        ou_accounts_cache=None,
        max_resources_per_template=None):
    base_template = copy.deepcopy(template)

    if "Transform" in base_template:
        if base_template["Transform"] == TRANSFORM_NAME:
            del base_template["Transform"]
        else:
            base_template["Transform"] = [t for t in base_template["Transform"] if t != TRANSFORM_NAME]

    resource_keys = [k for k, v in base_template["Resources"].items() if v["Type"] == RESOURCE_TYPE]
    resource_dict = {k: base_template["Resources"].pop(k) for k in resource_keys}
    LOGGER.debug(f"Found AssignmentGroup resources: {', '.join(resource_keys)}")

    configs = {}

    for resource_name, resource in resource_dict.items():
        validate_resource(resource)
        config = Config()
        config.load_resource_properties(resource["Properties"])
        config.resource_name_prefix = resource_name

        if not config.instance:
            config.instance = ids.instance_arn

        configs[resource_name] = config

    resource_collection_dict = {}
    max_stack_resources = 0
    for resource_name, config in configs.items():
        resource_collection = resources.get_resources_from_config(
            config,
            ou_fetcher=lambda ou, recursive: api_utils.get_accounts_for_ou(session, ou, recursive, cache=ou_accounts_cache))

        max_stack_resources += templates.get_max_number_of_child_stacks(resource_collection.num_resources, max_resources_per_template=max_resources_per_template)

        resource_collection_dict[resource_name] = resource_collection

    return base_template, max_stack_resources, resource_collection_dict


def handler(event, context, put_object=None):
    LOGGER.setLevel(getattr(logging, os.environ.get("LOG_LEVEL", "INFO")))
    logging.basicConfig()

    MAX_RESOURCES_PER_TEMPLATE = int(os.environ["MAX_RESOURCES_PER_TEMPLATE"]) if os.environ.get("MAX_RESOURCES_PER_TEMPLATE") else None
    MAX_CONCURRENT_ASSIGNMENTS = int(os.environ["MAX_CONCURRENT_ASSIGNMENTS"]) if os.environ.get("MAX_CONCURRENT_ASSIGNMENTS") else None

    BUCKET_NAME = os.environ["BUCKET_NAME"]
    KEY_PREFIX = os.environ.get("KEY_PREFIX", "")

    try:
        S3_PUT_OBJECT_ARGS = json.loads(os.environ["S3_PUT_OBJECT_ARGS"]) if os.environ.get("S3_PUT_OBJECT_ARGS") else {}
    except:
        LOGGER.exception("Error parsing S3_PUT_OBJECT_ARGS")

    request_id = event["requestId"]
    LOGGER.info(f"Request ID: {request_id}")

    input_template = event["fragment"]

    LOGGER.debug(f"Input template:\n{utils.dump_yaml(input_template)}")

    if "Resources" not in input_template:
        raise TypeError(f"{TRANSFORM_NAME} can only be used as a template-level transform")

    ou_accounts_cache = {}

    LOGGER.info("Extracting resources from template")
    output_template, max_stack_resources, resource_collection_dict = process_template(input_template,
            session=SESSION,
            ids=IDS,
            ou_accounts_cache=ou_accounts_cache,
            max_resources_per_template=MAX_RESOURCES_PER_TEMPLATE)

    all_child_templates_to_write = []

    s3_base_path_parts = [
        "templates",
        f"{request_id}",
    ]
    if KEY_PREFIX:
        s3_base_path_parts.insert(1, KEY_PREFIX)
    s3_base_path = "/".join(s3_base_path_parts)

    LOGGER.info(f"Processing {len(resource_collection_dict)} resources")
    for resource_name, resource_collection in resource_collection_dict.items():
        num_parent_resources = len(output_template["Resources"]) + max_stack_resources

        parent_template = templates.resolve_templates(
            resource_collection.assignments,
            resource_collection.permission_sets,
            max_resources_per_template=MAX_RESOURCES_PER_TEMPLATE,
            num_parent_resources=num_parent_resources,
        )

        parent_template_to_write, child_templates_to_write = parent_template.get_templates(
            s3_base_path,
            f"https://s3.amazonaws.com/{BUCKET_NAME}/{s3_base_path}",
            resource_name,
            ".yaml",
            base_template=output_template,
            max_concurrent_assignments=MAX_CONCURRENT_ASSIGNMENTS,
            path_joiner=lambda *args: "/".join(args)

        )
        output_template = parent_template_to_write.template
        LOGGER.debug(f"Intermediate output template:\n{utils.dump_yaml(output_template)}")

        all_child_templates_to_write.extend(child_templates_to_write)

    LOGGER.info(f"Writing {len(all_child_templates_to_write)} child templates")

    bucket = boto3.resource("s3").Bucket(BUCKET_NAME)

    for child_template_key, child_template in all_child_templates_to_write:
        LOGGER.debug(f"Writing child template {child_template_key}")
        content = utils.dump_yaml(child_template)

        put_object_args = copy.deepcopy(S3_PUT_OBJECT_ARGS)

        put_object_args.update({
            "Key": child_template_key,
            "Body": content,
        })
        if "ContentType" not in put_object_args:
            put_object_args["ContentType"] = "text/plain"

        if put_object:
            put_object(**put_object_args)
        else:
            bucket.put_object(**put_object_args)

    output_template = cfn_yaml_tags.to_json(output_template)

    LOGGER.debug(f"Final output template:\n{utils.dump_yaml(output_template)}")

    output = {
        "requestId" : request_id,
        "status" : "success",
        "fragment" : output_template,
    }

    return output




# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import logging
import re
from collections import namedtuple

import botocore
from botocore.exceptions import ProfileNotFound

LOGGER = logging.getLogger(__name__)

def _get(names):
    for name in names:
        value = os.environ.get(name)
        if value:
            return value, name
    return None, None

START_URL_VARS = ["AWS_DEFAULT_SSO_START_URL"]
REGION_VARS    = ["AWS_DEFAULT_SSO_REGION"]

class SSOInstance(namedtuple("SSOInstance", ["start_url", "region", "start_url_source", "region_source"])):
    def to_str(self, region=None):
        if region is None:
            region_str = "[NO_REGION]" if not self.region else ""
        elif region:
            region_str = f"[{self.region}]"
        else:
            region_str = ""
        return f"{self.start_url}{region_str}"

    def __str__(self):
        return self.to_str()

    def __bool__(self):
        return bool(self.start_url) and bool(self.region)

    @classmethod
    def to_strs(cls, instances, region=None):
        return ", ".join(i.to_str(region=region) for i in instances)
def _get_instance_from_profile(profile_name, scoped_config: dict, missing_ok=False) -> SSOInstance:
    start_url = scoped_config.get("sso_start_url")
    region = scoped_config.get("sso_region")
    if not (start_url and region):
        if not missing_ok:
            LOGGER.debug(f"Did not find config in profile {profile_name}")
        return None
    instance = SSOInstance(start_url, region, "profile", "profile")
    LOGGER.debug(f"Found instance {instance} in profile {profile_name}")
    return instance

def _get_all_instances_from_config(full_config: dict):
    instances = {}
    for profile_name, scoped_config in full_config.get("profiles", {}).items():
        instance = _get_instance_from_profile(profile_name, scoped_config, missing_ok=True)
        if not instance:
            continue
        if instance.start_url in instances and instance.region != instances[instance.start_url].region:
            regions = f"{instance.region}, {instances[instance.start_url].region}"
            LOGGER.warning(f"Region mismatch in config for {instance.start_url}: {regions}")
        else:
            instances[instance.start_url] = instance

    return sorted(instances.values(), key=lambda i: i.start_url)

def _validate_instance(instance: SSOInstance, source, start_url, start_url_source, region, region_source):
    if start_url and instance.start_url != start_url:
        message = f"start URL {instance.start_url} does not match {start_url} from {start_url_source}"
        LOGGER.warn(message)
    if region and instance.region != region:
        message = f"region {instance.region} does not match {region} from {region_source}"
        LOGGER.warn(message)

def _find_instance_from_profile(
        profile_name=None,
        profile_source=None,
        start_url=None,
        start_url_source=None,
        region=None,
        region_source=None):
    try:
        session = botocore.session.Session(profile=profile_name)
        instance = _get_instance_from_profile(profile_name, session.get_scoped_config())
    except ProfileNotFound:
        return None
    if not instance:
        return None
    _validate_instance(instance, profile_source, start_url, start_url_source, region, region_source)
    return instance

def _get_specifier(
        start_url=None,
        start_url_source=None,
        region=None,
        region_source=None,
        start_url_vars=None,
        region_vars=None):

    if start_url is None and region is None:
        if start_url_vars:
            start_url, start_url_source = _get(start_url_vars)
        if region_vars:
            region, region_source = _get(region_vars)

    if start_url is None and region is None:
        start_url, start_url_source = _get(START_URL_VARS)
        region, region_source = _get(REGION_VARS)

    return SSOInstance(start_url, region, start_url_source, region_source)

def _specifier_matches(specifier, instance):
    if specifier.start_url:
        if specifier.start_url.startswith("http"):
            if not specifier.start_url == instance.start_url:
                return False
        elif not re.search(specifier.start_url, instance.start_url):
            return False
        if specifier.region and specifier.region != instance.region:
            LOGGER.warn(
                f"Instance {instance.start_url} " +
                f"from {instance.start_url_source} " +
                f"has region {instance.region} " +
                f"which does not match {specifier.region} " +
                f"from {specifier.region_source}"
            )

    if specifier.region:
        if specifier.region != instance.region:
            return False

    return True


def find_instances(
        profile_name=None,
        profile_source=None,
        start_url=None,
        start_url_source=None,
        region=None,
        region_source=None,
        start_url_vars=None,
        region_vars=None):
    if profile_name:
        instance = _find_instance_from_profile()
        if instance:
            return [instance], None, [instance]
        else:
            return [], None, []

    specifier = _get_specifier(
        start_url=start_url,
        start_url_source=start_url_source,
        region=region,
        region_source=region_source,
        start_url_vars=start_url_vars,
        region_vars=region_vars,
    )

    if specifier:
        parts = [
            f"Using AWS SSO instance spec"
        ]
        if specifier.start_url:
            parts.extend([
                f"{specifier.start_url}",
                f"from {specifier.start_url_source}"
            ])
        if specifier.region:
            if specifier.start_url:
                parts.append("with")
            parts.extend([
                "region",
                f"{specifier.region}",
                f"from {specifier.region_source}"
            ])
        LOGGER.debug(" ".join(parts))

    if specifier and specifier.start_url.startswith("http"):
        return [specifier], specifier, [specifier]

    session = botocore.session.Session()
    all_instances = _get_all_instances_from_config(session.full_config)

    if not (specifier.start_url or specifier.region):
        return all_instances, specifier, all_instances

    instances = []
    for instance in all_instances:
        match = _specifier_matches(specifier, instance)
        if match:
            instances.append(instance)

    return instances, specifier, all_instances



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numbers
import jsonschema
import logging

from . import utils, cfn_yaml_tags
from .. import api_utils

LOGGER = logging.getLogger(__name__)

class ConfigError(Exception):
    pass

def _get_value(dct, keys, ensure_list=False):
    if not isinstance(keys, list):
        keys = [keys]
    for key in keys:
        if key in dct:
            value = dct[key]
            if ensure_list and not isinstance(value, list):
                value = [value]
            return key, value
    if ensure_list:
        return None, []
    else:
        return None, None

class Config:
    def __init__(self, data=None, resource_properties=None, resource_name_prefix=None):
        self.instance = None

        self.groups = []
        self.users = []

        self.permission_sets = []

        self.ous = []
        self.recursive_ous = []
        self.accounts = []

        self.resource_name_prefix = resource_name_prefix

        if data:
            self.load(data)
        if resource_properties:
            self.load_resource_properties(resource_properties)

    def load(self, data):
        for name in ["Instance", "InstanceArn", "InstanceARN"]:
            if name in data:
                self.instance = data[name]
                break

        def get(names):
            for name in names:
                if name in data:
                    value = data[name]
                    if not isinstance(value, list):
                        value = [value]
                    return value
            return []

        self.groups.extend(get(["Groups", "Group"]))
        self.users.extend(get(["Users", "User"]))

        self.permission_sets.extend(get(["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"]))

        self.ous.extend(get(["OUs", "Ous", "OU", "Ou"]))
        self.recursive_ous.extend(get(["RecursiveOUs", "RecursiveOus", "RecursiveOU", "RecursiveOu"]))
        for account in get(["Accounts", "Account"]):
            if isinstance(account, (str, numbers.Number)):
                account = str(int(account)).rjust(12, '0')
            self.accounts.append(account)


    def load_resource_properties(self, resource_properties):
        data = {}
        _, instance = _get_value(resource_properties, ["Instance", "InstanceArn", "InstanceARN"])
        if instance is not None:
            data["Instance"] = instance

        _, principals = _get_value(resource_properties, ["Principal", "Principals"], ensure_list=True)
        for principal_entry in principals:
            _, principal_type = _get_value(principal_entry, ["Type", "PrincipalType"])
            _, principal_ids = _get_value(principal_entry, ["Id", "PrincipalId", "Ids", "PrincipalIds"], ensure_list=True)
            if principal_type.upper() == "GROUP":
                config_key = "Groups"
            elif principal_type.upper() == "USER":
                config_key = "Users"
            else:
                raise ValueError(f"Invalid principal type: {principal_type}")
            if config_key not in data:
                data[config_key] = []
            data[config_key].extend(principal_ids)

        _, permission_sets = _get_value(resource_properties, ["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"], ensure_list=True)
        data["PermissionSets"] = permission_sets

        _, targets = _get_value(resource_properties, ["Target", "Targets"], ensure_list=True)
        for target_entry in targets:
            _, target_type = _get_value(target_entry, ["Type", "TargetType"])
            _, target_ids = _get_value(target_entry, ["Id", "TargetId", "Ids", "TargetIds"], ensure_list=True)
            if target_type.upper() == "AWS_OU":
                if target_entry.get("Recursive", False):
                    config_key = "RecursiveOus"
                else:
                    config_key = "Ous"
            elif target_type.upper() == "AWS_ACCOUNT":
                config_key = "Accounts"
            else:
                raise ValueError(f"Invalid target type: {target_type}")
            if config_key not in data:
                data[config_key] = []
            data[config_key].extend(target_ids)

        import yaml

        self.load(data)

FUNC_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "Ref|Fn::.*": {},
    },
    "minProperties": 1,
    "maxProperties": 1,
}

def _opt_func(schema):
    return {
        "oneOf": [
            schema,
            FUNC_SCHEMA,
        ]
    }

def _opt_list(schema, func=False):
    if func:
        return {
            "oneOf": [
                schema,
                {
                    "type": "array",
                    "items": {
                        "oneOf": [
                            schema,
                            FUNC_SCHEMA,
                        ]
                    }
                },
                FUNC_SCHEMA,
            ]
        }
    else:
        return {
            "oneOf": [
                schema,
                {
                    "type": "array",
                    "items": schema,
                },
            ]
        }

PRINCIPAL_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "(Principal)?Type": _opt_func({
            "type": "string",
            "enum": ["GROUP", "USER"],
        }),
        "(Principal)?Id(s)?": _opt_list({"type": "string"}, func=True),
    }
}

PERMISSION_SET_SCHEMA = _opt_func({
    "type": "string",
})

TARGET_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "(Target)?Type": _opt_func({
            "type": "string",
            "enum": ["AWS_OU", "AWS_ACCOUNT"],
        }),
        "(Target)?Id(s)?": _opt_list({"type": ["string", "integer"]}, func=True),
        "Recursive": {"type": "boolean"},
    }
}

RESOURCE_PROPERTY_SCHEMA = {
    "type": "object",
    "patternProperties": {
        "Name": _opt_func({"type": "string"}),
        "Instance(Arn|ARN)?": _opt_func({"type": "string"}),
        "Principal(s)?": _opt_list(PRINCIPAL_SCHEMA),
        "PermissionSet(Arn)?(s)?": _opt_list(PERMISSION_SET_SCHEMA),
        "Target(s)?": _opt_list(TARGET_SCHEMA),
        "UpdateNonce": _opt_func({"type": "string"}),
    },
    "additionalProperties": False,
}

def _check(properties, keys, required=True, parent="Resource"):
    found_keys = [k for k in keys if k in properties]
    if len(found_keys) > 1:
        raise ConfigError(f"{parent} must have only one of {', '.join(found_keys)}")
    if required and len(found_keys) == 0:
        raise ConfigError(f"{parent} must have one of {', '.join(keys)}")

def validate_resource(resource):
    resource = cfn_yaml_tags.to_json(resource)
    properties = resource.get("Properties", {})
    try:
        jsonschema.validate(
            schema=RESOURCE_PROPERTY_SCHEMA,
            instance=properties)
    except jsonschema.ValidationError as e:
        raise ConfigError(f"Resource is invalid: {e!s}")

    _check(properties, ["Instance", "InstanceArn", "InstanceARN"], required=False)

    _check(properties, ["Principal", "Principals"])

    name, principals = _get_value(properties, ["Principal", "Principals"], ensure_list=True)
    for principal_entry in principals:
        _check(principal_entry, ["Type", "PrincipalType"], parent=name)
        _check(principal_entry, ["Id", "PrincipalId", "Ids", "PrincipalIds"], parent=name)

    _check(properties, ["PermissionSet", "PermissionSetArn", "PermissionSets", "PermissionSetArns"])

    name, targets = _get_value(properties, ["Target", "Targets"], ensure_list=True)
    for target_entry in targets:
        _check(target_entry, ["Type", "TargetType"], parent=name)
        _check(target_entry, ["Id", "TargetId", "Ids", "TargetIds"], parent=name)

def validate_config(config, ids: api_utils.Ids):
    if not config.instance:
        config.instance = ids.instance_arn
    elif not ids.instance_arn_matches(config.instance):
        LOGGER.warning(f"Config instance {config.instance} does not match {ids.instance_arn}")

    if not (config.groups or config.users):
       raise ConfigError("No principals specified")
    if not config.permission_sets:
        raise ConfigError("No permission sets specified")
    if not (config.ous or config.accounts):
        raise ConfigError("No targets specified")



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
import sys
from collections import namedtuple
import logging

import boto3

import click

from ..api_utils import Ids
from .utils import configure_logging

LOGGER = logging.getLogger(__name__)

@click.command('lookup')
@click.argument('type', type=click.Choice(['instance', 'identity-store', 'group', 'user', 'permission-set']))
@click.argument('value', nargs=-1)

@click.option('--instance-arn', '--ins')
@click.option('--identity-store-id', '--ids')

@click.option('--profile')

@click.option('--error-if-not-found', '-e', is_flag=True)
@click.option('--show-id/--hide-id', default=False, help='Print SSO instance/identity store id being used')
@click.option('--separator', '--sep', default=': ')
@click.option('--verbose', count=True)
def lookup(
        type,
        value,
        instance_arn,
        identity_store_id,
        profile,
        error_if_not_found,
        show_id,
        separator,
        verbose):
    configure_logging(LOGGER, verbose)

    session = boto3.Session(profile_name=profile)

    ids = Ids(lambda: session, instance_arn, identity_store_id)
    ids.suppress_print = not show_id

    try:
        if type == 'instance':
            ids.suppress_print = True
            print(ids.instance_arn)
        elif type == 'identity-store':
            ids.suppress_print = True
            print(ids.identity_store_id)
        elif type in 'group':
            if not value:
                raise click.UsageError("Group name is required")
            lines = []
            for name in value:
                try:
                    group_id = lookup_group_by_name(session, ids, name)
                except LookupError as e:
                    if error_if_not_found:
                        print(format_lines(lines, separator))
                        print("Group {} not found".format(name), file=sys.stderr)
                        sys.exit(1)
                    group_id = 'NOT_FOUND'
                lines.append((name, group_id))
            print(format_lines(lines, separator))
        elif type == 'user':
            if not value:
                raise click.UsageError("User name is required")
            lines = []
            for name in value:
                try:
                    user_id = lookup_user_by_name(session, ids, name)
                except LookupError as e:
                    if error_if_not_found:
                        print(format_lines(lines, separator))
                        print("User {} not found".format(name), file=sys.stderr)
                        sys.exit(1)
                    user_id = 'NOT_FOUND'
                lines.append((name, user_id))
            print(format_lines(lines, separator))
        elif type == 'permission-set':
            if not value:
                raise click.UsageError("Permission set name is required")
            lookup = PermissionSetArnLookup(session, ids)
            lines = []
            for name in value:
                try:
                    permission_set_arn = lookup.lookup_permission_set_arn(name)
                except LookupError as e:
                    if error_if_not_found:
                        print(format_lines(lines, separator))
                        print("Permission set {} not found".format(name), file=sys.stderr)
                        sys.exit(1)
                    permission_set_arn = 'NOT_FOUND'
                lines.append((name, permission_set_arn))
            print(format_lines(lines, separator))

    except LookupError as e:
        print(e, file=sys.stderr)
        sys.exit(1)

def format_lines(lines, separator):
    max_len = max(len(l[0]) for l in lines)
    return '\n'.join("{}{}{}".format(l[0].ljust(max_len), separator, l[1]) for l in lines)

def lookup_group_by_name(session, ids, name):
    identity_store_client = session.client('identitystore')
    filters=[{'AttributePath': 'DisplayName', 'AttributeValue': name}]
    try:
        response = identity_store_client.list_groups(IdentityStoreId=ids.identity_store_id, Filters=filters)
        if len(response['Groups']) == 0:
            raise LookupError("No group named {} found".format(name))
        elif len(response['Groups']) > 1:
            raise LookupError("{} groups named {} found".format(len(response['Groups']), name))
        return response['Groups'][0]['GroupId']
    except:
        raise

def lookup_user_by_name(session, ids, name):
    identity_store_client = session.client('identitystore')
    filters=[{'AttributePath': 'UserName', 'AttributeValue': name}]
    try:
        response = identity_store_client.list_users(IdentityStoreId=ids.identity_store_id, Filters=filters)
        if len(response['Users']) == 0:
            raise LookupError("No user named {} found".format(name))
        elif len(response['Users']) > 1:
            raise LookupError("{} users named {} found".format(len(response['Users']), name))
        return response['Users'][0]['UserId']
    except:
        raise

class PermissionSetArnLookup:
    def __init__(self, session, ids):
        self.client = session.client('sso-admin')
        self.paginator = self.client.get_paginator('list_permission_sets')
        self.instance_arn = ids.instance_arn
        self.cache = {}

    def lookup_permission_set_arn(self, name):
        if name in self.cache:
            return self.cache[name]
        for response in self.paginator.paginate(InstanceArn=self.instance_arn):
            for permission_set_arn in response['PermissionSets']:
                ps_description = self.client.describe_permission_set(InstanceArn=self.instance_arn, PermissionSetArn=permission_set_arn)
                self.cache[ps_description['PermissionSet']['Name']] = permission_set_arn
            if name in self.cache:
                return self.cache[name]
        raise LookupError("No permission set named {} found".format(name))

if __name__ == '__main__':
    lookup(prog_name="python -m aws_sso_util.cli.lookup")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# modified from code at
# https://github.com/aws/aws-cli/blob/v2/awscli/customizations/sso/logout.py

import json
import logging
import os

import botocore
from botocore.exceptions import ClientError

from ..sso import SSO_TOKEN_DIR
from ..sso import CREDENTIALS_CACHE_DIR as AWS_CREDS_CACHE_DIR

from .utils import configure_logging

import click

LOGGER = logging.getLogger(__name__)

@click.command()
@click.option("--debug", is_flag=True)
def logout(debug):
    configure_logging(LOGGER, debug)
    session = botocore.session.Session()

    LOGGER.debug("Removing tokens")
    SSOTokenSweeper(session).delete_credentials(SSO_TOKEN_DIR)
    LOGGER.debug("Removing credentials")
    SSOCredentialSweeper().delete_credentials(AWS_CREDS_CACHE_DIR)


class BaseCredentialSweeper(object):
    def delete_credentials(self, creds_dir):
        if not os.path.isdir(creds_dir):
            return
        filenames = os.listdir(creds_dir)
        for filename in filenames:
            filepath = os.path.join(creds_dir, filename)
            contents = self._get_json_contents(filepath)
            if contents is None:
                continue
            if self._should_delete(contents):
                self._before_deletion(contents)
                os.remove(filepath)

    def _should_delete(self, filename):
        raise NotImplementedError('_should_delete')

    def _get_json_contents(self, filename):
        try:
            with open(filename, 'r') as f:
                return json.load(f)
        except Exception:
            # We do not want to include the traceback in the exception
            # so that we do not accidentally log sensitive contents because
            # of the exception or its Traceback.
            LOGGER.debug('Failed to load: %s', filename)
            return None

    def _before_deletion(self, contents):
        pass


class SSOTokenSweeper(BaseCredentialSweeper):
    def __init__(self, session):
        self._session = session

    def _should_delete(self, contents):
        return 'accessToken' in contents

    def _before_deletion(self, contents):
        # If the sso region is present in the cached token, construct a client
        # and invoke the logout api to invalidate the token before deleting it.
        sso_region = contents.get('region')
        if sso_region:
            config = botocore.config.Config(
                region_name=sso_region,
                signature_version=botocore.UNSIGNED,
            )
            sso = self._session.create_client('sso', config=config)
            try:
                sso.logout(accessToken=contents['accessToken'])
            except ClientError:
                # The token may alread be expired or otherwise invalid. If we
                # get a client error on logout just log and continue on
                LOGGER.debug('Failed to call logout API:', exc_info=True)


class SSOCredentialSweeper(BaseCredentialSweeper):
    def _should_delete(self, contents):
        return contents.get('ProviderType') == 'sso'


if __name__ == "__main__":
    logout(prog_name="python -m aws_sso_util.cli.logout")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import re
from collections import namedtuple

import boto3
import aws_error_utils

import click

from ..api_utils import Ids

def get_principal_filter(group_values, user_values):
    def filter(type, id, name):
        if not (group_values or user_values):
            return True
        if type == "GROUP":
            for value in group_values:
                if id == value or re.search(value, name):
                    return True
        elif type == "USER":
            for value in user_values:
                if id == value or re.search(value, name):
                    return True
        else:
            raise ValueError(f"Unknown principal type {type}")
    return filter

def get_permission_set_filter(values):
    def filter(arn, name):
        if not values:
            return True
        for value in values:
            if arn == value:
                return True
            if arn.split("/", 1)[1] == value:
                return True
            if arn.split("/", 2)[2] == value:
                return True
            if re.search(value, name):
                return True
        return False
    return filter

def get_target_filter(values):
    def filter(type, id, name):
        if type != "AWS_ACCOUNT":
            return True
        if not values:
            return True
        for value in values:
            if id.startswith(value) or id.endswith(value) or re.search(value, name):
                return True
        return False
    return filter

@click.command()
@click.option("--instance-arn", "--ins")
@click.option("--identity-store-id", "--ids")

@click.option("--group", "-g", "group_values", multiple=True, default=[])
@click.option("--user", "-u", "user_values", multiple=True, default=[])
@click.option("--permission-set", "-p", "permission_set_values", multiple=True, default=[])
@click.option("--account", "-a", "account_values", multiple=True, default=[])

@click.option('--show-id/--hide-id', default=False, help='Print SSO instance/identity store id being used')
@click.option('--separator', '--sep', default=',')
def assignments(
        instance_arn,
        identity_store_id,
        group_values,
        user_values,
        permission_set_values,
        account_values,
        show_id,
        separator):
    session = boto3.Session()

    ids = Ids(lambda: session, instance_arn, identity_store_id)
    ids.suppress_print = not show_id

    principal_filter = get_principal_filter(group_values, user_values)
    permission_set_filter = get_permission_set_filter(permission_set_values)
    target_filter = get_target_filter(account_values)

    for assignment in lookup_assignments(session, ids, principal_filter, permission_set_filter, target_filter):
        print(f"{separator}".join(assignment))

Assignment = namedtuple("Assignment", [
    "principal_type",
    "principal_id",
    "principal_name",
    "permission_set_id",
    "permission_set_name",
    "target_type",
    "target_id",
    "target_name",
])

def lookup_assignments(session, ids: Ids, principal_filter, permission_set_filter, target_filter):
    organizations_client = session.client("organizations")
    sso_admin_client = session.client("sso-admin")
    identity_store_client = session.client("identitystore")

    cache = {}
    filter_cache = {}

    accounts_paginator = organizations_client.get_paginator("list_accounts")
    for response in accounts_paginator.paginate():
        for account in response["Accounts"]:
            account_id = account["Id"]
            account_name = account["Name"]

            if account_id not in filter_cache:
                filter_cache[account_id] = target_filter("AWS_ACCOUNT", account_id, account_name)
            if not filter_cache[account_id]:
                continue

            permission_sets_paginator = sso_admin_client.get_paginator("list_permission_sets_provisioned_to_account")
            for response in permission_sets_paginator.paginate(
                    InstanceArn=ids.instance_arn,
                    AccountId=account_id):
                if "PermissionSets" not in response:
                    continue

                for permission_set_arn in response["PermissionSets"]:
                    if permission_set_arn not in cache:
                        response = sso_admin_client.describe_permission_set(
                            InstanceArn=ids.instance_arn,
                            PermissionSetArn=permission_set_arn
                        )
                        cache[permission_set_arn] = response["PermissionSet"]["Name"]
                    permission_set_name = cache[permission_set_arn]
                    permission_set_id = permission_set_arn.split("/", 1)[1]

                    if permission_set_arn not in filter_cache:
                        filter_cache[permission_set_arn] = permission_set_filter(permission_set_arn, permission_set_name)
                    if not filter_cache[permission_set_arn]:
                        continue

                    assignments_paginator = sso_admin_client.get_paginator("list_account_assignments")

                    for response in assignments_paginator.paginate(
                            InstanceArn=ids.instance_arn,
                            AccountId=account_id,
                            PermissionSetArn=permission_set_arn):
                        for assignment in response["AccountAssignments"]:
                            principal_type = assignment["PrincipalType"]
                            principal_id = assignment["PrincipalId"]
                            principal_key = (principal_type, principal_id)
                            if principal_key not in cache:
                                if principal_type == "GROUP":
                                    try:
                                        response = identity_store_client.describe_group(
                                            IdentityStoreId=ids.identity_store_id,
                                            GroupId=principal_id
                                        )
                                        cache[principal_key] = response["DisplayName"]
                                    except aws_error_utils.catch_aws_error("ResourceNotFoundException"):
                                        cache[principal_key] = "UNKNOWN"
                                elif principal_type == "USER":
                                    try:
                                        response = identity_store_client.describe_user(
                                            IdentityStoreId=ids.identity_store_id,
                                            UserId=principal_id
                                        )
                                        cache[principal_key] = response["UserName"]
                                    except aws_error_utils.catch_aws_error("ResourceNotFoundException"):
                                        cache[principal_key] = "UNKNOWN"
                                else:
                                    raise ValueError(f"Unknown principal type {principal_type}")
                            principal_name = cache[principal_key]

                            if principal_key not in filter_cache:
                                filter_cache[principal_key] = principal_filter(principal_type, principal_id, principal_name)
                            if not filter_cache[principal_key]:
                                continue

                            yield (
                                principal_type,
                                principal_id,
                                principal_name,
                                permission_set_id,
                                permission_set_name,
                                "AWS_ACCOUNT",
                                account_id,
                                account_name,
                            )

if __name__ == "__main__":
    assignments(prog_name="python -m aws_sso_util.cli.assignments")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is based on the code for the AWS CLI v2"s `aws sso login` functionality
# https://github.com/aws/aws-cli/tree/v2/awscli/customizations/sso

import argparse
import os
import sys
import json
import logging
import datetime

from botocore.session import Session
from botocore.exceptions import ClientError

import click

from ..sso import get_token_loader, get_credentials
from ..exceptions import InvalidSSOConfigError, AuthDispatchError, AuthenticationNeededError
from ..vendored_botocore.exceptions import UnauthorizedSSOTokenError

LOG_FILE = os.path.expanduser(
    os.path.join("~", ".aws", "sso", "aws-sso-credential-process-log.txt")
)

LOGGER = logging.getLogger(__name__)

CONFIG_VARS = [
    ("start url", "sso_start_url"),
    ("SSO region", "sso_region"),
    ("account", "sso_account_id"),
    ("role", "sso_role_name")
]

def get_config(arg_config, profile_config):
    sso_config = {}
    missing_vars = []
    for friendly_name, config_var_name in CONFIG_VARS:
        if arg_config.get(config_var_name):
            sso_config[config_var_name] = arg_config[config_var_name]
        elif config_var_name not in profile_config:
            missing_vars.append((friendly_name, config_var_name))
            sso_config[config_var_name] = None
        else:
            sso_config[config_var_name] = profile_config[config_var_name]

    required_vars = ["sso_start_url", "sso_region", "sso_account_id", "sso_role_name"]

    missing_requred_vars = [v[0] for v in missing_vars if v[1] in required_vars]
    if missing_requred_vars:
        raise InvalidSSOConfigError(
            "Missing " + ", ".join(missing_requred_vars)
        )
    return sso_config


@click.command("credential-process")
@click.option("--profile", help="Extract settings from the given profile")
@click.option("--sso-start-url", "--start-url", "start_url")
@click.option("--sso-region", "--region", "region")
@click.option("--account-id")
@click.option("--role-name")

@click.option("--force-refresh", is_flag=True, help="Do not reuse cached AWS SSO token")
@click.option( "--verbose", "--debug", count=True, help="Write to the debugging log file")
def credential_process(
        profile,
        start_url,
        region,
        account_id,
        role_name,
        force_refresh,
        verbose):

    if verbose or os.environ.get("AWS_SSO_CREDENTIAL_PROCESS_DEBUG", "").lower() in ["1", "true"]:
        logging.basicConfig(level=logging.DEBUG, filename=LOG_FILE, filemode="w")
    else:
        logging.disable(logging.CRITICAL)

    LOGGER.info("Starting credential process at {}".format(datetime.datetime.now().isoformat()))

    if role_name is None and os.environ.get("AWS_SSO_ROLE_NAME"):
        LOGGER.debug("Using role from env: {}".format(os.environ["AWS_SSO_ROLE_NAME"]))
        role_name = os.environ["AWS_SSO_ROLE_NAME"]

    if account_id is None and os.environ.get("AWS_SSO_ACCOUNT_ID"):
        LOGGER.debug("Using acccount from env: {}".format(os.environ["AWS_SSO_ACCOUNT_ID"]))
        account_id = os.environ["AWS_SSO_ACCOUNT_ID"]

    # if role_name and role_name.startswith("arn"):
    #     parts = role_name.split(":")
    #     account_id = parts[4]
    #     role_name = parts[5].split("/", 1)[1]

    if start_url is None and os.environ.get("AWS_SSO_START_URL"):
        start_url = os.environ["AWS_SSO_START_URL"]

    if region is None and os.environ.get("AWS_SSO_REGION"):
        region = os.environ["AWS_SSO_REGION"]

    session_kwargs = {}

    if profile:
        session_kwargs["profile"] = profile

    arg_config = {
        "sso_start_url": start_url,
        "sso_region": region,
        "sso_role_name": role_name,
        "sso_account_id": account_id,
    }

    LOGGER.info("CONFIG FROM ARGS: {}".format(json.dumps(arg_config)))

    try:
        session = Session(**session_kwargs)

        if profile:
            profile_config = session.get_scoped_config()
            LOGGER.info("CONFIG FROM PROFILE: {}".format(json.dumps(profile_config)))
        else:
            profile_config = {}

        config = get_config(arg_config, profile_config)

        LOGGER.info("CONFIG: {}".format(json.dumps(config)))

        if (config.get("sso_interactive_auth") or "").lower() == "true":
            raise InvalidSSOConfigError("Interactive auth has been removed. See https://github.com/benkehoe/aws-sso-credential-process/issues/4")

        token_loader = get_token_loader(
            session=session,
            sso_region=config["sso_region"],
            force_refresh=force_refresh,
        )

        if not config["sso_account_id"]:
            raise InvalidSSOConfigError("Missing account id")

        if not config["sso_role_name"]:
            raise InvalidSSOConfigError("Missing role")

        credentials = get_credentials(
            session=session,
            start_url=config["sso_start_url"],
            sso_region=config["sso_region"],
            account_id=config["sso_account_id"],
            role_name=config["sso_role_name"],
            token_loader=token_loader,
        )

        output = {
            "Version": 1,
            "AccessKeyId": credentials["access_key"],
            "SecretAccessKey": credentials["secret_key"],
            "SessionToken": credentials["token"],
            # as provided the expiration isn"t valid ISO8601 and that causes parsing errors for some SDKs
            "Expiration": credentials["expiry_time"].replace("UTC", "Z"),
        }
        LOGGER.debug("CREDENTIALS: " + json.dumps(output))

        print(json.dumps(output, separators=(",", ":")))
    except (AuthenticationNeededError, UnauthorizedSSOTokenError) as e:
        if profile:
            aws_sso_util_cmd = f"aws-sso-util login --profile {profile}"
            aws_sso_cmd = f"aws sso login --profile {profile}"
        else:
            aws_sso_util_cmd = f"aws-sso-util login {config['sso_start_url']} {config['sso_region']}"
            aws_sso_cmd = f"aws sso login"
        print(f"Login required. Use `{aws_sso_util_cmd}` or `{aws_sso_cmd}` and try again.", file=sys.stderr)
        sys.exit(1)
    except InvalidSSOConfigError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(2)
    except AuthDispatchError as e:
        LOGGER.error(e)
        print(e, file=sys.stderr)
        sys.exit(3)
    except ClientError as e:
        LOGGER.error(e, exc_info=True)
        #TODO: print a different message for AccessDeniedException during CreateToken? -> user canceled login
        # boto_error_matches(e, "CreateToken", "AccessDeniedException")
        print("ERROR:", e, file=sys.stderr)
        sys.exit(4)
    except Exception as e:
        LOGGER.error(e, exc_info=True)
        print("ERROR:", e, file=sys.stderr)
        sys.exit(5)

if __name__ == "__main__":
    credential_process(prog_name="python -m aws_sso_util.cli.credential_process")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2018 iRobot Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements support for CloudFormation's YAML tags.
Does not add these to the "safe" methods by default.
Call the mark_safe() method to add it to these methods.

import json
import yaml
import cfn_yaml_tags
cfn_yaml_tags.mark_safe()

with open('template.yaml', 'r') as fp:
    template = yaml.safe_load(fp)

template["Resources"]["MyResource"]["Properties"]["Foo"] = cfn_yaml_tags.Ref("Bar")

json.dumps(template, cls=cfn_yaml_tags.JSONEncoder)

plain_template = cfn_yaml_tags.to_json(template)
json.dumps(plain_template)

"""

__version__ = '1.1.0'

import six
import itertools
import re
import json

import yaml
from yaml.constructor import ConstructorError

class CloudFormationObject(object):
    SCALAR = 'scalar'
    SEQUENCE = 'sequence'
    SEQUENCE_OR_SCALAR = 'sequence_scalar'
    MAPPING = 'mapping'
    MAPPING_OR_SCALAR = 'mapping_scalar'

    name = None
    tag = None
    type = None

    def __init__(self, data):
        self.data = data

    def to_json(self):
        """Return the JSON equivalent"""
        def convert(obj):
            return obj.to_json() if isinstance(obj, CloudFormationObject) else obj

        if isinstance(self.data, dict):
            data = type(self.data)((key, convert(value)) for key, value in six.iteritems(self.data))
        elif isinstance(self.data, (list, tuple)):
            data = type(self.data)(convert(value) for value in self.data)
        else:
            data = self.data

        name = self.name

        if name == 'Fn::GetAtt' and isinstance(data, six.string_types):
            data = data.split('.')
        elif name == 'Ref' and '.' in data:
            name = 'Fn::GetAtt'
            data = data.split('.')

        return {name: data}

    @classmethod
    def construct(cls, loader, node):
        if cls.type == cls.SCALAR:
            return cls(loader.construct_scalar(node))
        elif cls.type == cls.SEQUENCE:
            return cls(loader.construct_sequence(node))
        elif cls.type == cls.SEQUENCE_OR_SCALAR:
            try:
                return cls(loader.construct_sequence(node))
            except ConstructorError:
                return cls(loader.construct_scalar(node))
        elif cls.type == cls.MAPPING:
            return cls(loader.construct_mapping(node))
        elif cls.type == cls.MAPPING_OR_SCALAR:
            try:
                return cls(loader.construct_mapping(node))
            except ConstructorError:
                return cls(loader.construct_scalar(node))
        else:
            raise RuntimeError("Unknown type {}".format(cls.type))

    @classmethod
    def represent(cls, dumper, obj):
        data = obj.data
        if isinstance(data, list):
            return dumper.represent_sequence(obj.tag, data)
        elif isinstance(data, dict):
            return dumper.represent_mapping(obj.tag, data)
        else:
            return dumper.represent_scalar(obj.tag, data)

    def __str__(self):
        return '{} {}'.format(self.tag, self.data)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.data))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and other.data == self.data

def to_json(obj):
    """Return the JSON equivalent"""
    if isinstance(obj, CloudFormationObject):
        return obj.to_json()

    if isinstance(obj, dict):
        obj = type(obj)((k, to_json(v)) for k, v in obj.items())
    elif isinstance(obj, (list, tuple)):
        obj = type(obj)(to_json(v) for v in obj)

    return obj

class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, CloudFormationObject):
            return o.to_json()
        return json.JSONEncoder.default(self, o)

# (name, tag, type)

ref = ('Ref', 'Ref', CloudFormationObject.SCALAR)

functions = [
    ('Fn::And',         'And',         CloudFormationObject.SEQUENCE),
    ('Fn::Condition',   'Condition',   CloudFormationObject.SCALAR),
    ('Fn::Base64',      'Base64',      CloudFormationObject.SCALAR),
    ('Fn::Equals',      'Equals',      CloudFormationObject.SEQUENCE),
    ('Fn::FindInMap',   'FindInMap',   CloudFormationObject.SEQUENCE),
    ('Fn::GetAtt',      'GetAtt',      CloudFormationObject.SEQUENCE_OR_SCALAR),
    ('Fn::GetAZs',      'GetAZs',      CloudFormationObject.SCALAR),
    ('Fn::If',          'If',          CloudFormationObject.SEQUENCE),
    ('Fn::ImportValue', 'ImportValue', CloudFormationObject.SCALAR),
    ('Fn::Join',        'Join',        CloudFormationObject.SEQUENCE),
    ('Fn::Not',         'Not',         CloudFormationObject.SEQUENCE),
    ('Fn::Or',          'Or',          CloudFormationObject.SEQUENCE),
    ('Fn::Select',      'Select',      CloudFormationObject.SEQUENCE),
    ('Fn::Split',       'Split',       CloudFormationObject.SEQUENCE),
    ('Fn::Sub',         'Sub',         CloudFormationObject.SEQUENCE_OR_SCALAR),
]

_object_classes = None

def init(safe=False):
    global _object_classes
    _object_classes = []
    for name_, tag_, type_ in itertools.chain(functions, [ref]):
        if not tag_.startswith('!'):
            tag_ = '!{}'.format(tag_)
        tag_ = six.u(tag_)

        class Object(CloudFormationObject):
            name = name_
            tag = tag_
            type = type_
        obj_cls_name = re.search(r'\w+$', tag_).group(0)
        if six.PY2:
            obj_cls_name = str(obj_cls_name)
        Object.__name__ = obj_cls_name

        _object_classes.append(Object)
        globals()[obj_cls_name] = Object

        yaml.add_constructor(tag_, Object.construct)
        yaml.add_representer(Object, Object.represent)

    if safe:
        mark_safe()

def mark_safe():
    for obj_cls in _object_classes:
        yaml.add_constructor(obj_cls.tag, obj_cls.construct, Loader=yaml.SafeLoader)
        yaml.add_representer(obj_cls, obj_cls.represent, Dumper=yaml.SafeDumper)

init()



# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# this is the new content of botocore.credentials in the v2 branch

import time
import datetime
import logging
import os
import getpass
import threading
import json
import subprocess
from collections import namedtuple
from copy import deepcopy
from hashlib import sha1

from dateutil.parser import parse
from dateutil.tz import tzlocal, tzutc

import botocore.configloader
import botocore.compat
from botocore import UNSIGNED
# from botocore.compat import total_seconds
# from botocore.compat import compat_shell_split
from botocore.config import Config
# from botocore.exceptions import UnknownCredentialError
# from botocore.exceptions import PartialCredentialsError
# from botocore.exceptions import ConfigNotFound
# from botocore.exceptions import InvalidConfigError
# from botocore.exceptions import InfiniteLoopConfigError
# from botocore.exceptions import RefreshWithMFAUnsupportedError
# from botocore.exceptions import MetadataRetrievalError
# from botocore.exceptions import CredentialRetrievalError
# from botocore.exceptions import UnauthorizedSSOTokenError
# from botocore.utils import InstanceMetadataFetcher, parse_key_val_file
# from botocore.utils import ContainerMetadataFetcher
# from botocore.utils import FileWebIdentityTokenLoader
# from botocore.utils import SSOTokenLoader

from botocore.credentials import (
    CachedCredentialFetcher,
    _serialize_if_needed,
)

from .exceptions import UnauthorizedSSOTokenError


class SSOCredentialFetcher(CachedCredentialFetcher):
    def __init__(self, start_url, sso_region, role_name, account_id,
                 client_creator, token_loader=None, cache=None,
                 expiry_window_seconds=None):
        self._client_creator = client_creator
        self._sso_region = sso_region
        self._role_name = role_name
        self._account_id = account_id
        self._start_url = start_url
        self._token_loader = token_loader

        super(SSOCredentialFetcher, self).__init__(
            cache, expiry_window_seconds
        )

    def _create_cache_key(self):
        """Create a predictable cache key for the current configuration.
        The cache key is intended to be compatible with file names.
        """
        args = {
            'startUrl': self._start_url,
            'roleName': self._role_name,
            'accountId': self._account_id,
        }
        # NOTE: It would be good to hoist this cache key construction logic
        # into the CachedCredentialFetcher class as we should be consistent.
        # Unfortunately, the current assume role fetchers that sub class don't
        # pass separators resulting in non-minified JSON. In the long term,
        # all fetchers should use the below caching scheme.
        args = json.dumps(args, sort_keys=True, separators=(',', ':'))
        argument_hash = sha1(args.encode('utf-8')).hexdigest()
        return self._make_file_safe(argument_hash)

    def _parse_timestamp(self, timestamp_ms):
        # fromtimestamp expects seconds so: milliseconds / 1000 = seconds
        timestamp_seconds = timestamp_ms / 1000.0
        timestamp = datetime.datetime.fromtimestamp(timestamp_seconds, tzutc())
        return _serialize_if_needed(timestamp)

    def _get_credentials(self):
        """Get credentials by calling SSO get role credentials."""
        config = Config(
            signature_version=UNSIGNED,
            region_name=self._sso_region,
        )
        client = self._client_creator('sso', config=config)

        kwargs = {
            'roleName': self._role_name,
            'accountId': self._account_id,
            'accessToken': self._token_loader(self._start_url),
        }
        try:
            response = client.get_role_credentials(**kwargs)
        except client.exceptions.UnauthorizedException:
            raise UnauthorizedSSOTokenError()
        credentials = response['roleCredentials']

        credentials = {
            'ProviderType': 'sso',
            'Credentials': {
                'AccessKeyId': credentials['accessKeyId'],
                'SecretAccessKey': credentials['secretAccessKey'],
                'SessionToken': credentials['sessionToken'],
                'Expiration': self._parse_timestamp(credentials['expiration']),
            }
        }
        return credentials



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging

import boto3

LOGGER = logging.getLogger(__name__)

class RetrievalError(Exception):
    pass

def get_accounts_for_ou(session, ou, recursive, refresh=False, cache=None):
    if not cache:
        cache = {}

    ou_children_key = f"{ou}#children"
    ou_accounts_key = f"{ou}#accounts"

    accounts = []

    if refresh or ou_accounts_key not in cache:
        LOGGER.info(f"Retrieving accounts for OU {ou}")
        client = session.client('organizations')

        ou_accounts = []

        paginator = client.get_paginator('list_accounts_for_parent')
        for response in paginator.paginate(ParentId=ou):
            ou_accounts.extend(data['Id'] for data in response['Accounts'])

        cache[ou_accounts_key] = ou_accounts

        accounts.extend(ou_accounts)
    else:
        accounts.extend(cache[ou_accounts_key])

    if recursive:
        if refresh or ou_children_key not in cache:
            LOGGER.info(f"Retrieving child OUs for OU {ou}")
            client = session.client('organizations')

            ou_children = []

            paginator = client.get_paginator('list_organizational_units_for_parent')
            for response in paginator.paginate(ParentId=ou):
                sub_ous = [data['Id'] for data in response['OrganizationalUnits']]
                ou_children.extend(sub_ous)

            cache[ou_children_key] = ou_children

        for sub_ou_id in cache[ou_children_key]:
            accounts.extend(get_accounts_for_ou(session, sub_ou_id, True, refresh=refresh, cache=cache))

    return accounts

class LookupError(Exception):
    pass

class Ids:
    def __init__(self, session_fetcher, instance_arn=None, identity_store_id=None):
        if instance_arn and not instance_arn.startswith('arn:'):
            instance_arn = f"arn:aws:sso:::instance/{instance_arn}"

        self._session_fetcher = session_fetcher
        self._client = None
        self._instance_arn = instance_arn
        self._instance_arn_printed = False
        self._identity_store_id = identity_store_id
        self._identity_store_id_printed = False
        self.suppress_print = False

    def _get_client(self):
        if not self._client:
            self._client = self._session_fetcher().client('sso-admin')
        return self._client

    def _print(self, *args, **kwargs):
        if not self.suppress_print:
            print(*args, **kwargs)

    def instance_arn_matches(self, instance):
        if not self._instance_arn:
            return True
        if instance and not instance.startswith('arn:'):
            instance = f"arn:aws:sso:::instance/{instance}"
        return instance == self._instance_arn

    @property
    def instance_arn(self):
        if self._instance_arn:
            if not self._instance_arn_printed:
                self._print("Using SSO instance {}".format(self._instance_arn.split('/')[-1]))
                self._instance_arn_printed = True
            return self._instance_arn
        response = self._get_client().list_instances()
        if len(response['Instances']) == 0:
            raise LookupError("No SSO instance found, please specify with --instance-arn")
        elif len(response['Instances']) > 1:
            raise LookupError("{} SSO instances found, please specify with --instance-arn".format(len(response['Instances'])))
        else:
            instance_arn = response['Instances'][0]['InstanceArn']
            self._instance_arn = instance_arn
            self._print("Using SSO instance {}".format(self._instance_arn.split('/')[-1]))
            self._instance_arn_printed = True
            identity_store_id = response['Instances'][0]['IdentityStoreId']
            if self._identity_store_id and self._identity_store_id != identity_store_id:
                raise LookupError("SSO instance identity store {} does not match given identity store {}".format(identity_store_id, self._identity_store_id))
            else:
                self._identity_store_id = identity_store_id
        return self._instance_arn

    @property
    def identity_store_id(self):
        if self._identity_store_id:
            if not self._identity_store_id_printed:
                self._print("Using SSO identity store {}".format(self._identity_store_id))
                self._identity_store_id_printed = True
            return self._identity_store_id
        response = self._get_client().list_instances()
        if len(response['Instances']) == 0:
            raise LookupError("No SSO instance found, please specify identity store with --identity-store-id or instance with --instance-arn")
        elif len(response['Instances']) > 1:
            raise LookupError("{} SSO instances found, please specify identity store with --identity-store-id or instance with --instance-arn".format(len(response['Instances'])))
        else:
            identity_store_id = response['Instances'][0]['IdentityStoreId']
            self._identity_store_id = identity_store_id
            self._print("Using SSO identity store {}".format(self._identity_store_id))
            self._identity_store_id_printed = True
            instance_arn = response['Instances'][0]['InstanceArn']
            instance_id = instance_arn.split('/')[-1]
            if self._instance_arn and self._instance_arn != instance_arn:
                raise LookupError("SSO instance {} does not match given instance {}".format(instance_id, self._instance_arn.split('/')[-1]))
            else:
                self._instance_arn = instance_arn
        return self._identity_store_id



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import re
import math
from collections import OrderedDict, namedtuple
from pathlib import PurePath
import logging

from . import resources
from . import utils

LOGGER = logging.getLogger(__name__)

MAX_RESOURCES_PER_TEMPLATE = 500
MAX_CONCURRENT_ASSIGNMENTS = 20

WritableTemplate = namedtuple("WritableTemplate", ["path", "template"])
TemplateCollection = namedtuple("TemplateCollection", [
    "parent",
    "children",
])
_ChildData = namedtuple("_ChildData", ["path_for_writing", "path_for_resource", "stem", "template"])

def is_name_in_template(name, template):
    for section in ["Parameters", "Conditions", "Resources"]:
        if name in template.get(section, {}):
            return True
    return False

def add_parameters_to_template(template,
        base_template: dict=None,
        template_parameters: dict=None,
        references: set=None):
    if "Parameters" not in template:
        template["Parameters"] = OrderedDict()

    if base_template and "Parameters" in base_template:
        template["Parameters"].update(base_template["Parameters"])

    if template_parameters:
        for parameter_name, default in template_parameters:
            if parameter_name not in template["Parameters"]:
                template["Parameters"][parameter_name] = OrderedDict({"Type": "String"})
            if default is not None:
                template["Parameters"][parameter_name]["Default"] = default

    if references:
        for reference_name in sorted(references):
            if not is_name_in_template(reference_name, template):
                template["Parameters"][reference_name] = OrderedDict({"Type": "String"})

    if not template["Parameters"]:
        del template["Parameters"]

def add_assignments_to_template(
        template,
        assignments: resources.AssignmentResources,
        child_stack,
        max_concurrent_assignments=None):
    max_concurrent_assignments = _get_num(max_concurrent_assignments, MAX_CONCURRENT_ASSIGNMENTS)

    assignment_resources = []
    for assignment in assignments:
        if max_concurrent_assignments > 0 and len(assignment_resources) >= max_concurrent_assignments:
            depends_on = assignment_resources[len(assignment_resources)-max_concurrent_assignments][0]
        else:
            depends_on = None

        assignment_resources.append((
            assignment.get_resource_name(),
            assignment.get_resource(child_stack=child_stack, depends_on=depends_on)))

    if "Resources" not in template:
        template["Resources"] = OrderedDict()

    template["Resources"].update(assignment_resources)

class ChildTemplate:
    def __init__(self, assignments: resources.AssignmentResources):
        self.assignments = assignments

    def get_references(self):
        return self.assignments.references

    def get_template(self,
            max_concurrent_assignments=None,
            resource_name_prefix=None):
        template = OrderedDict({
            "AWSTemplateFormatVersion": "2010-09-09",
        })
        add_parameters_to_template(template, references=sorted(self.get_references()))
        template["Resources"] = OrderedDict(template.get("Resources", {}))

        add_assignments_to_template(template, self.assignments,
                child_stack=True,
                max_concurrent_assignments=max_concurrent_assignments)

        return template

class ParentTemplate:
    def __init__(self,
            assignments: resources.AssignmentResources=None,
            permission_sets: resources.PermissionSetResources=None,
            child_templates=None):
        self.assignments = assignments
        self.permission_sets = permission_sets
        self.child_templates = list(child_templates) if child_templates else []

    def _get_template(self,
            child_templates=None,
            base_template=None,
            parameters=None,
            max_concurrent_assignments=None):
        template = OrderedDict({
            "AWSTemplateFormatVersion": "2010-09-09",
        })

        references = set()
        for child in self.child_templates:
            references.update(child.get_references())
        for assignment in self.assignments:
            references.update(assignment.references)

        for permission_set in self.permission_sets:
            references.discard(permission_set.get_resource_name())

        if base_template:
            found_references = set()
            for name in references:
                if is_name_in_template(name, base_template):
                    found_references.add(name)
            references -= found_references

        add_parameters_to_template(template,
                base_template=base_template,
                template_parameters=parameters,
                references=references)

        if base_template:
            for key in base_template:
                if key in ["AWSTemplateFormatVersion", "Parameters"]:
                    continue
                template[key] = utils.to_ordered_dict(base_template[key])

        if "Resources" not in template:
            template["Resources"] = OrderedDict()

        if self.permission_sets:
            for permission_set in self.permission_sets:
                resource_name = permission_set.get_resource_name()
                if not resource_name:
                    continue
                template["Resources"][resource_name] = permission_set.get_resource()

        if self.assignments:
            add_assignments_to_template(template, self.assignments,
                    child_stack=False,
                    max_concurrent_assignments=max_concurrent_assignments)

        def get_reference(name):
            DATA = [
                (("AWS::SSO::PermissionSet"), "PermissionSetArn"),
            ]
            for resource_types, attr in DATA:
                if name in template["Resources"] and template["Resources"][name]["Type"] in resource_types:
                    if attr:
                        return utils.GETATT_TAG([name, attr])
            return utils.REF_TAG(name)

        if child_templates:
            child_resource_names = []
            for child in child_templates:
                if max_concurrent_assignments and child_resource_names:
                    depends_on = child_resource_names[-1]
                else:
                    depends_on = None

                resource_name = re.sub(r'[^a-zA-Z0-9]', '', child.stem)
                resource = OrderedDict({
                    "Type": "AWS::CloudFormation::Stack",
                })
                if depends_on:
                    resource["DependsOn"] = depends_on
                resource["Properties"] =  OrderedDict({
                    "TemplateURL": str(child.path_for_resource)
                })
                if "Parameters" in child.template and child.template["Parameters"]:
                    resource["Properties"]["Parameters"] = OrderedDict()
                    for parameter_name in child.template["Parameters"].keys():
                        resource["Properties"]["Parameters"][parameter_name] = get_reference(parameter_name)
                template["Resources"][resource_name] = resource

                child_resource_names.append(resource_name)

        return template

    def get_templates(self,
            base_path: str,
            child_base_path_for_resource: str,
            stem,
            template_file_suffix,
            base_template=None,
            parameters=None,
            max_concurrent_assignments=None,
            child_templates_in_subdir=True,
            path_joiner=None):
        if not path_joiner:
            path_joiner = os.path.join

        child_templates = []

        for i, child in enumerate(self.child_templates):
            child_path_for_writing = base_path
            child_path_for_resource = child_base_path_for_resource

            if child_templates_in_subdir:
                child_path_for_writing = path_joiner(child_path_for_writing, stem)
                child_path_for_resource = path_joiner(child_path_for_resource, stem)
            child_stem = f"{stem}{i:03d}"
            child_path_for_writing = path_joiner(child_path_for_writing, f"{child_stem}{template_file_suffix}")
            child_path_for_resource = path_joiner(child_path_for_resource, f"{child_stem}{template_file_suffix}")
            child_templates.append(_ChildData(
                child_path_for_writing,
                child_path_for_resource,
                child_stem,
                child.get_template(
                    max_concurrent_assignments=max_concurrent_assignments)))

        parent_path = path_joiner(base_path, f"{stem}{template_file_suffix}")
        parent_template = self._get_template(
            child_templates=child_templates,
            base_template=base_template,
            parameters=parameters,
            max_concurrent_assignments=max_concurrent_assignments)

        return TemplateCollection(
            parent=WritableTemplate(parent_path, parent_template),
            children=[WritableTemplate(c.path_for_writing, c.template) for c in child_templates]
        )

def resolve_templates(
        assignments: resources.AssignmentResources,
        permission_sets: resources.PermissionSetResources,
        max_resources_per_template: int=None,
        num_parent_resources: int=0) -> ParentTemplate:

    max_resources_per_template = _get_num(max_resources_per_template, MAX_RESOURCES_PER_TEMPLATE)
    if assignments.num_resources() + permission_sets.num_resources() + num_parent_resources > max_resources_per_template:
        child_templates = [ChildTemplate(c) for c in assignments.chunk(max_resources_per_template)]
        parent_assignments = resources.AssignmentResources([])
    else:
        parent_assignments = assignments
        child_templates = []

    parent_template = ParentTemplate(parent_assignments,
            permission_sets=permission_sets,
            child_templates=child_templates)

    return parent_template

def get_max_number_of_child_stacks(num_resources, max_resources_per_template=None):
    max_per_template = _get_num(max_resources_per_template, MAX_RESOURCES_PER_TEMPLATE)

    return math.ceil(num_resources/max_per_template)

def _get_num(value, default):
    if value is None or value > 0:
        return value
    else:
        return default



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import json
from collections import OrderedDict

import yaml
from . import cfn_yaml_tags
cfn_yaml_tags.mark_safe()

def to_ordered_dict(obj):
    if isinstance(obj, dict):
        return OrderedDict((k, to_ordered_dict(v)) for k, v in obj.items())
    elif isinstance(obj, list):
        return [to_ordered_dict(v) for v in obj]
    else:
        return obj

def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

yaml.add_representer(OrderedDict, represent_ordereddict, Dumper=yaml.SafeDumper)
yaml.SafeDumper.ignore_aliases = lambda *args : True

def load_yaml(*args, **kwargs):
    return yaml.safe_load(*args, **kwargs)

def dump_yaml(*args, **kwargs):
    return yaml.safe_dump(*args, **kwargs)

def get_instance_id_from_arn(instance_arn):
    return instance_arn.split('/', 1)[1]

REF_TAG = getattr(cfn_yaml_tags, "Ref")
GETATT_TAG = getattr(cfn_yaml_tags, "GetAtt")
def get_references(value):
    references = set()
    if isinstance(value, REF_TAG):
        references.add(value.data)
    elif isinstance(value, dict) and "Ref" in value and len(value) == 1:
        references.add(value["Ref"].split(".")[0])
    elif isinstance(value, dict) and "Fn::GetAtt" in value and len(value) == 1:
        get_att_value = value["Fn::GetAtt"]
        if isinstance(get_att_value, str):
            references.add(get_att_value.split(".")[0])
        else:
            references.add(get_att_value[0])
    elif isinstance(value, cfn_yaml_tags.CloudFormationObject):
        value_json = value.to_json()
        references.update(get_references(value_json))
    elif isinstance(value, (list, set)):
        for v in value:
            references.update(get_references(v))
    elif isinstance(value, dict):
        for v in value.values():
            references.update(get_references(v))
    return references

def is_reference(value):
    if isinstance(value, cfn_yaml_tags.CloudFormationObject):
        return True
    elif isinstance(value, dict) and "Ref" in value and len(value) == 1:
        return True
    elif isinstance(value, dict) and "Fn::GetAtt" in value and len(value) == 1:
        return True
    else:
        return False

def get_hash_key(value):
    if isinstance(value, cfn_yaml_tags.CloudFormationObject):
        value = value.to_json()
    if not isinstance(value, str):
        value = json.dumps(value, sort_keys=True)
    return value.encode('utf-8')

def chunk_list_generator(lst, chunk_length):
    for i in range(0, len(lst), chunk_length):
        yield lst[i:i + chunk_length]



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import enum
import logging
import hashlib
from collections import OrderedDict, namedtuple

from .config import Config
from . import utils
from . import cfn_yaml_tags

LOGGER = logging.getLogger(__name__)

class Principal:
    class Type(enum.Enum):
        GROUP = 'GROUP'
        USER = 'USER'

    def __init__(self, type: 'Principal.Type', name):
        self.type = type
        self.name = name

        self.references = utils.get_references(self.name)

    @property
    def hash_key(self):
        return f"{self.type.value}:{utils.get_hash_key(self.name)}".encode('utf-8')

    def __str__(self):
        return f"{self.type.name}:{self.name}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.type!r}, {self.name!r})"

class PermissionSet:
    RESOURCE_NAME_PREFIX = "PermSet"

    class _Type(enum.Enum):
        ARN = 'ARN'
        SHORT_ARN = 'SHORT_ARN'
        ID = 'ID'
        REFERENCE = 'REF'
        RESOURCE = 'RESOURCE'

    @classmethod
    def _get_type(cls, value):
        if utils.is_reference(value):
            return cls._Type.REFERENCE
        if isinstance(value, dict):
            return cls._Type.RESOURCE
        if isinstance(value, str) and value.startswith('arn'):
            return cls._Type.ARN
        if isinstance(value, str) and (value.startswith('ssoins-') or value.startswith('ins-')):
            return cls._Type.SHORT_ARN
        if isinstance(value, str) and value.startswith('ps-'):
            return cls._Type.ID
        raise TypeError(f"Unknown permission set type {type(value)} for {value}")

    def __init__(self, value, instance, resource_name_prefix=None):
        self._instance = instance

        self._type = self._get_type(value)
        self._value = value
        self._arn = None

        self._resource_name_prefix = resource_name_prefix

        self.references = utils.get_references(self._value)
        if self._type == self._Type.RESOURCE:
            self.references.add(self.get_resource_name())

    @property
    def hash_key(self):
        return utils.get_hash_key(self.get_arn())

    def get_arn(self, force_ref=False):
        if self._type == self._Type.RESOURCE:
            if force_ref:
                utils.REF_TAG(self.get_resource_name())
            else:
                return utils.GETATT_TAG(f"{self.get_resource_name()}.PermissionSetArn")
        if self._type == self._Type.REFERENCE:
            if force_ref:
                references = utils.get_references(self._value)
                if len(references) != 1:
                    raise ValueError(f"Cannot convert to ref: {self._value}")
                return utils.REF_TAG(list(references)[0])
            return self._value
        if self._type == self._Type.ARN:
            return self._value
        if self._type == self._Type.SHORT_ARN:
            return f"arn:aws:sso:::permissionSet/{self._value}"
        if self._type == self._Type.ID:
            instance_id = utils.get_instance_id_from_arn(self._instance)
            return f"arn:aws:sso:::permissionSet/{instance_id}/{self._value}"
        raise TypeError(f"Invalid PermissionSet type {self._type}")

    def get_resource_name(self):
        prefix = self._resource_name_prefix or ''
        if self._type == self._Type.RESOURCE:
            return f"{prefix}{self.RESOURCE_NAME_PREFIX}{self._get_name()}"
        else:
            return None

    def get_resource(self):
        if self._type != self._Type.RESOURCE:
            return {}

        if "Type" in self._value:
            resource = self._value
        else:
            properties = OrderedDict()
            if "InstanceArn" not in self._value:
                properties["InstanceArn"] = self._instance
            properties.update(self._value)
            resource = OrderedDict({
                "Type": "AWS::SSO::PermissionSet",
                "Properties": properties
            })
        return resource

    def _get_name(self):
        if self._type != self._Type.RESOURCE:
            raise TypeError("PermissionSet._get_name() called when not a resource")
        if "Type" in self._value:
            return self._value["Properties"]["Name"]
        else:
            return self._value["Name"]

    def __str__(self):
        if self._type in [self._Type.ARN, self._Type.SHORT_ARN, self._Type.ID]:
            return self._value
        if self._type == self._Type.REFERENCE:
            return str(self._value)
        if self._type == self._Type.RESOURCE:
            return f"{{{self._get_name()}}}"
        raise TypeError(f"Invalid PermissionSet type {self._type}")

    def __repr__(self):
        return f"{self.__class__.__name__}({self._value!r})"

class Target:
    class Type(enum.Enum):
        OU = 'AWS_OU'
        ACCOUNT = 'AWS_ACCOUNT'

    def __init__(self, type: 'Target.Type', name: str, source_ou=None):
        self.type = type
        self.name = name
        self.source_ou = None

        self.references = utils.get_references(self.name)

    @property
    def hash_key(self):
        return f"{self.type.value}:{utils.get_hash_key(self.name)}".encode('utf-8')

    def __str__(self):
        s = f"{self.type.name}:{self.name}"
        if self.source_ou:
            s += f"[{self.source_ou}]"
        return s

    def __repr__(self):
        if self.source_ou:
            ou_str = f", source_ou={self.source_ou!r}"
        else:
            ou_str = ""
        return f"{self.__class__.__name__}({self.type!r}, {self.name!r}{ou_str})"

class Assignment:
    RESOURCE_NAME_PREFIX = "Assignment"

    def __init__(self, instance, principal, permission_set, target, resource_name_prefix=None):
        self.instance = instance
        self.principal = principal
        self.permission_set = permission_set
        self.target = target

        self._resource_name_prefix = resource_name_prefix

        self.references = utils.get_references(self.instance) | self.principal.references | self.permission_set.references | self.target.references

    def get_resource_name(self):
        prefix = self._resource_name_prefix or ''
        hasher = hashlib.md5()
        hasher.update(utils.get_hash_key(self.instance))
        hasher.update(self.principal.hash_key)
        hasher.update(self.permission_set.hash_key)
        hasher.update(self.target.hash_key)
        hash_value = hasher.hexdigest()[:6].upper()
        return f"{prefix}{self.RESOURCE_NAME_PREFIX}{hash_value}"

    def get_resource(self, child_stack, depends_on=None):
        resource = OrderedDict({
            "Type": "AWS::SSO::Assignment"
        })
        if self.target.source_ou:
            resource["Metadata"] = OrderedDict({
                "AccountSourceOU": self.target.source_ou
            })
        if depends_on:
            resource["DependsOn"] = depends_on
        resource["Properties"] = OrderedDict({
            "InstanceArn": self.instance,
            "PrincipalType": self.principal.type.value,
            "PrincipalId": self.principal.name,
            "PermissionSetArn": self.permission_set.get_arn(force_ref=child_stack),
            "TargetType": self.target.type.value,
            "TargetId": self.target.name
        })

        return resource

    def __str__(self):
        return f"<{self.principal!s}|{self.permission_set!s}|{self.target!s}>"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.instance!r}, {self.principal!r}, {self.permission_set!r}, {self.target!r})"

class ResourceList:
    def __init__(self, resources):
        self._resources = list(resources)
        self.references = set()

        self._init(resources)

    def _init(self, resources):
        for resource in resources:
            self.references.update(resource.references)

    def chunk(self, max_resources):
        return [self.__class__(chunk) for chunk in utils.chunk_list_generator(self._resources, max_resources)]

    def __str__(self):
        return f"[{', '.join(str(v) for v in self._resources)}]"

    def __repr__(self):
        return f"{self.__class__.__name__}({self._resources!r})"

    def __len__(self):
        return len(self._resources)

    def __iter__(self):
        return iter(self._resources)

    def extend(self, resources):
        if isinstance(resources, ResourceList):
            resources = resources._resources
        self._resources.extend(resources)
        self._init(resources)

class AssignmentResources(ResourceList):
    def __init__(self, assignments):
        super().__init__(assignments)

    def num_resources(self):
        return len(self._resources)

class PermissionSetResources(ResourceList):
    def __init__(self, permission_sets):
        super().__init__(permission_sets)

    def num_resources(self):
        return len([r for r in self._resources if r._type == PermissionSet._Type.RESOURCE])

ResourceCollection = namedtuple("ResourceCollection", ["num_resources", "assignments", "permission_sets"])

def get_resources_from_config(config: Config, ou_fetcher=None) -> ResourceCollection:
    if config.instance is None:
        raise ValueError("SSO instance is not set on config")

    num_resources = 0

    principals = []
    for group in config.groups:
        LOGGER.debug(f"Group: {group!s} {group!r}")
        principals.append(Principal(Principal.Type.GROUP, group))
    for user in config.users:
        principals.append(Principal(Principal.Type.USER, user))
    LOGGER.debug(f"Got principals: [{', '.join(str(v) for v in principals)}]")

    permission_sets = [
        PermissionSet(ps, instance=config.instance, resource_name_prefix=config.resource_name_prefix)
        for ps in config.permission_sets
    ]
    LOGGER.debug(f"Got permission sets: [{', '.join(str(v) for v in permission_sets)}]")

    targets = []

    if (config.ous or config.recursive_ous) and not ou_fetcher:
        LOGGER.error("OU specified but ou_fetcher not provided")
        raise ValueError("OU specified but ou_fetcher not provided")

    for ou in config.ous:
        LOGGER.debug(f"Translating OU {ou} to accounts")
        accounts = ou_fetcher(ou, recursive=False)

        for account in accounts:
            targets.append(Target(Target.Type.ACCOUNT, account, source_ou=ou))

    for ou in config.recursive_ous:
        LOGGER.debug(f"Translating OU {ou} recursively to accounts")
        accounts = ou_fetcher(ou, recursive=True)

        for account in accounts:
            targets.append(Target(Target.Type.ACCOUNT, account, source_ou=ou))

        for account in accounts:
            targets.append(Target(Target.Type.ACCOUNT, account, source_ou=ou))

    for account in config.accounts:
        targets.append(Target(Target.Type.ACCOUNT, account))
    LOGGER.debug(f"Got targets: [{', '.join(str(v) for v in targets)}]")

    assignments = []
    for principal in principals:
        for permission_set in permission_sets:
            for target in targets:
                assignments.append(Assignment(
                    config.instance,
                    principal,
                    permission_set,
                    target,
                    resource_name_prefix=config.resource_name_prefix,
                ))
    LOGGER.debug(f"Got assignments: [{', '.join(str(v) for v in assignments)}]")

    ar = AssignmentResources(assignments)
    psr = PermissionSetResources(permission_sets)
    num_resources = ar.num_resources() + psr.num_resources()

    return ResourceCollection(num_resources, ar, psr)



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
import subprocess
import tempfile
import sys
from pathlib import Path
import shutil
import shlex
import contextlib

import click

REPO = "https://github.com/benkehoe/aws-sso-util"

def clone_repo(git_clone_origin, git_clone_branch, git_clone_args):
    temp_dir = tempfile.TemporaryDirectory(prefix="aws-sso-util")

    temp_dir_path = Path(temp_dir.name)

    git_command = ["git", "clone", git_clone_origin, str(temp_dir_path)]
    if git_clone_branch:
        git_command.extend(["--branch", git_clone_branch])
    else:
        #git_command.extend(["--branch", "macro/stable"])
        pass
    if git_clone_args:
        git_command.extend(shlex.split(git_clone_args))
    else:
        git_command.extend(["--depth", "1"])


    print(f"Running git cline: {shlex.join(git_command)}")
    result = subprocess.run(git_command)

    if result.returncode:
        print("git clone failed", file=sys.stderr)
        sys.exit(2)

    subprocess.run(f"ls -al {temp_dir_path}", shell=True)

    return temp_dir

@click.command("deploy-macro")
@click.option("--load-samconfig", help="Use an existing samconfig.toml file")
@click.option("--save-samconfig", help="Save the resulting samconfig.toml to a given path, or 'true' to save to the loaded file")
@click.option("--sam-build-args", help="A string of arguments to pass to 'sam build'")
@click.option("--sam-deploy-args", default="--guided", help="A string of arguments to pass to 'sam deploy'")
@click.option("--git-clone-branch")
@click.option("--git-clone-args", help="A string of arguments to pass to 'git clone'")
@click.option("--git-clone-origin", default=REPO)
@click.option("--existing-repo-dir", help="Use an existing repo path instead of a temporary clone")
def deploy_macro(
        load_samconfig,
        save_samconfig,
        sam_build_args,
        sam_deploy_args,
        git_clone_branch,
        git_clone_args,
        git_clone_origin,
        existing_repo_dir):
    if existing_repo_dir:
        repo_dir = contextlib.nullcontext(existing_repo_dir)
    else:
        repo_dir = clone_repo(git_clone_origin, git_clone_branch, git_clone_args)

    with repo_dir as repo_dir_path:
        repo_dir_path = Path(repo_dir_path)
        working_dir = repo_dir_path / "macro"
        if not working_dir.is_dir():
            print(f"{working_dir} is not a valid directory", file=sys.stderr)
            sys.exit(1)

        if load_samconfig:
            load_samconfig = Path(load_samconfig)
            if not load_samconfig.is_file():
                print(f"{load_samconfig} is not a valid file", file=sys.stderr)
                sys.exit(1)
            print(f"Copying samconfig.toml from {load_samconfig} to {working_dir}")
            shutil.copy2(load_samconfig.resolve(), working_dir / "samconfig.toml")

        run_kwargs = {
            "cwd": working_dir
        }

        sam_build_command = ["sam", "build", "--use-container"]
        if sam_build_args:
            sam_build_command.extend(shlex.split(sam_build_args))

        print(f"Running sam build: {shlex.join(sam_build_command)}")
        result = subprocess.run(sam_build_command, **run_kwargs)

        if result.returncode:
            print("sam build failed", file=sys.stderr)
            sys.exit(2)

        sam_deploy_command = ["sam", "deploy"]
        if sam_deploy_args:
            sam_deploy_command.extend(shlex.split(sam_deploy_args))

        print(f"Running sam deploy: {shlex.join(sam_deploy_command)}")
        result = subprocess.run(sam_deploy_command, **run_kwargs)

        if result.returncode:
            print("sam deploy failed", file=sys.stderr)
            sys.exit(2)

        if save_samconfig:
            if save_samconfig.lower() in ["true", "1"]:
                if not load_samconfig:
                    print(f"--save-samconfig {save_samconfig} set but --load-samconfig not provided")
                    sys.exit(1)
                save_samconfig = load_samconfig
            save_samconfig = Path(save_samconfig)
            print(f"Writing samconfig.toml to {save_samconfig}")
            save_samconfig.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy2(working_dir / "samconfig.toml", save_samconfig)

if __name__ == "__main__":
    deploy_macro(prog_name="python -m aws_sso_util.cli.deploy_macro")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import logging.handlers

def configure_logging(logger, verbose, **config_args):
    if verbose in [False, None]:
        verbose = 0
    elif verbose == True:
        verbose = 1

    logging.basicConfig(**config_args)

    if verbose == 0:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(message)s"))
        logger.addHandler(handler)
        logger.propagate = False

    aws_sso_util_logger = logging.getLogger("aws_sso_util")
    if verbose >= 1:
        aws_sso_util_logger.setLevel(logging.DEBUG)
    else:
        aws_sso_util_logger.setLevel(logging.INFO)

    root_logger = logging.getLogger()
    if verbose >= 2:
        root_logger.setLevel(logging.DEBUG)




# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import click

from .. import __version__

from .assignments import assignments
from .cfn import generate_template
from .configure_profile import configure_profile
from .credential_process import credential_process
from .deploy_macro import deploy_macro
from .login import login
from .logout import logout
from .lookup import lookup
from .populate_profiles import populate_profiles

@click.group(name="aws-sso-util")
@click.version_option(version=__version__, message='%(version)s')
def cli():
    pass

@cli.group()
def configure():
    pass

configure.add_command(configure_profile, "profile")
configure.add_command(populate_profiles, "populate")

cli.add_command(login)
cli.add_command(logout)

cli.add_command(lookup)
cli.add_command(assignments)

# cli.add_command(deploy_macro)
cli.add_command(generate_template, "cfn")

cli.add_command(credential_process)




# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

# this is the new content of botocore.utils in the v2 branch

import time
import datetime
import logging
import os
import getpass
import threading
import json
import subprocess
from collections import namedtuple
from copy import deepcopy
from hashlib import sha1
import hashlib

from dateutil.parser import parse
from dateutil.tz import tzlocal
import dateutil

import botocore.configloader
import botocore.compat
from botocore import UNSIGNED
from botocore.compat import total_seconds
# from botocore.compat import compat_shell_split
from botocore.config import Config
# from botocore.exceptions import UnknownCredentialError
# from botocore.exceptions import PartialCredentialsError
# from botocore.exceptions import ConfigNotFound
# from botocore.exceptions import InvalidConfigError
# from botocore.exceptions import PendingAuthorizationExpiredError

from botocore.utils import (
    CachedProperty,
    datetime2timestamp,
    tzutc
)

from .exceptions import PendingAuthorizationExpiredError

class SSOTokenFetcher(object):
    # The device flow RFC defines the slow down delay to be an additional
    # 5 seconds:
    # https://tools.ietf.org/html/draft-ietf-oauth-device-flow-15#section-3.5
    _SLOW_DOWN_DELAY = 5
    # The default interval of 5 is also defined in the RFC (see above link)
    _DEFAULT_INTERVAL = 5
    _EXPIRY_WINDOW = 15 * 60
    _CLIENT_REGISTRATION_TYPE = 'public'
    _GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:device_code'

    def __init__(
            self, sso_region, client_creator, cache=None,
            on_pending_authorization=None, time_fetcher=None, sleep=None,
    ):
        self._sso_region = sso_region
        self._client_creator = client_creator
        self._on_pending_authorization = on_pending_authorization

        if time_fetcher is None:
            time_fetcher = self._utc_now
        self._time_fetcher = time_fetcher

        if sleep is None:
            sleep = time.sleep
        self._sleep = sleep

        if cache is None:
            cache = {}
        self._cache = cache

    def _utc_now(self):
        return datetime.datetime.now(tzutc())

    def _parse_if_needed(self, value):
        if isinstance(value, datetime.datetime):
            return value
        return dateutil.parser.parse(value)

    def _is_expired(self, response):
        end_time = self._parse_if_needed(response['expiresAt'])
        seconds = total_seconds(end_time - self._time_fetcher())
        return seconds < self._EXPIRY_WINDOW

    @CachedProperty
    def _client(self):
        config = botocore.config.Config(
            region_name=self._sso_region,
            signature_version=botocore.UNSIGNED,
        )
        return self._client_creator('sso-oidc', config=config)

    def _register_client(self):
        timestamp = datetime2timestamp(self._time_fetcher())
        response = self._client.register_client(
            # NOTE: As far as I know client name will never be returned to the
            # user. For now we'll just use botocore-client with the timestamp.
            clientName='botocore-client-%s' % int(timestamp),
            clientType=self._CLIENT_REGISTRATION_TYPE,
        )
        expires_at = response['clientSecretExpiresAt']
        expires_at = datetime.datetime.fromtimestamp(expires_at, tzutc())
        registration = {
            'clientId': response['clientId'],
            'clientSecret': response['clientSecret'],
            'expiresAt': expires_at,
        }
        return registration

    def _registration(self):
        # Registration with the OIDC endpoint is regional, is good for roughly
        # 90 days, and can be shared. This is currently scoped to individual
        # tools and as such each tool has their own cached client id.
        cache_key = 'botocore-client-id-%s' % self._sso_region
        if cache_key in self._cache:
            registration = self._cache[cache_key]
            if not self._is_expired(registration):
                return registration

        registration = self._register_client()
        self._cache[cache_key] = registration
        return registration

    def _authorize_client(self, start_url, registration):
        # NOTE: The authorization response is not cached. These responses are
        # short lived (currently only 10 minutes) and can only be exchanged for
        # a token once. Having multiple clients share this is problematic.
        response = self._client.start_device_authorization(
            clientId=registration['clientId'],
            clientSecret=registration['clientSecret'],
            startUrl=start_url,
        )
        expires_in = datetime.timedelta(seconds=response['expiresIn'])
        authorization = {
            'deviceCode': response['deviceCode'],
            'userCode': response['userCode'],
            'verificationUri': response['verificationUri'],
            'verificationUriComplete': response['verificationUriComplete'],
            'expiresAt': self._time_fetcher() + expires_in,
        }
        if 'interval' in response:
            authorization['interval'] = response['interval']
        return authorization

    def _poll_for_token(self, start_url):
        registration = self._registration()
        authorization = self._authorize_client(start_url, registration)

        if self._on_pending_authorization:
            # This callback can display the user code / verification URI
            # so the user knows the page to go to. Potentially, this call
            # back could even be used to auto open a browser.
            self._on_pending_authorization(**authorization)

        interval = authorization.get('interval', self._DEFAULT_INTERVAL)
        # NOTE: This loop currently relies on the service to either return
        # a valid token or a ExpiredTokenException to break the loop. If this
        # proves to be problematic it may be worth adding an additional
        # mechanism to control timing this loop out.
        while True:
            try:
                response = self._client.create_token(
                    grantType=self._GRANT_TYPE,
                    clientId=registration['clientId'],
                    clientSecret=registration['clientSecret'],
                    deviceCode=authorization['deviceCode'],
                )
                expires_in = datetime.timedelta(seconds=response['expiresIn'])
                token = {
                    'startUrl': start_url,
                    'region': self._sso_region,
                    'accessToken': response['accessToken'],
                    'expiresAt': self._time_fetcher() + expires_in
                }
                return token
            except self._client.exceptions.SlowDownException:
                interval += self._SLOW_DOWN_DELAY
            except self._client.exceptions.AuthorizationPendingException:
                pass
            except self._client.exceptions.ExpiredTokenException:
                raise PendingAuthorizationExpiredError()
            self._sleep(interval)

    def _token(self, start_url, force_refresh):
        cache_key = hashlib.sha1(start_url.encode('utf-8')).hexdigest()
        # Only obey the token cache if we are not forcing a refresh.
        if not force_refresh and cache_key in self._cache:
            token = self._cache[cache_key]
            if not self._is_expired(token):
                return token

        token = self._poll_for_token(start_url)
        self._cache[cache_key] = token
        return token

    def fetch_token(self, start_url, force_refresh=False):
        return self._token(start_url, force_refresh)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aws_sso_util',
 'aws_sso_util.cfn',
 'aws_sso_util.cli',
 'aws_sso_util.vendored_botocore']

package_data = \
{'': ['*']}

install_requires = \
['aws-error-utils>=1.0.4,<2.0.0',
 'boto3>=1.16.13,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'pyyaml>=5.3.1,<6.0.0']

entry_points = \
{'console_scripts': ['aws-sso-util = aws_sso_util.cli:cli']}

setup_kwargs = {
    'name': 'aws-sso-util',
    'version': '2.10.0',
    'description': 'Utilities to make AWS SSO easier',
    'long_description': "# aws-sso-util\n## Making life with AWS SSO a little easier\n\n[AWS SSO](https://aws.amazon.com/single-sign-on/) has some rough edges, and `aws-sso-util` is here to smooth them out, hopefully temporarily until AWS makes it better.\n\n`aws-sso-util` contains utilities for the following:\n* Configuring `.aws/config`\n* Logging in/out\n* AWS SDK support\n* Looking up identifiers\n* CloudFormation\n* Python library for AWS SSO authentication\n\n`aws-sso-util` supersedes `aws-sso-credential-process`, which is still available in its original form [here](https://github.com/benkehoe/aws-sso-credential-process).\nRead the updated docs for `aws-sso-util credential-process` [here](docs/credential-process.md).\n\n## Quickstart\n\n0. Make sure you've [installed the AWS CLI v2](https://docs.aws.amazon.com/cli/latest/userguide/install-cliv2.html) (which has AWS SSO support).\n\n1. I recommend you install [`pipx`](https://pipxproject.github.io/pipx/), which installs the tool in an isolated virtualenv while linking the script you need.\n\nMac:\n```bash\nbrew install pipx\npipx ensurepath\n```\n\nOther:\n```bash\npython3 -m pip install --user pipx\npython3 -m pipx ensurepath\n```\n\n2. Install\n```bash\npipx install aws-sso-util\n```\n\n3. Learn\n```bash\naws-sso-util --help\n```\n\n4. Autocomplete\n\n`aws-sso-util` uses [click](https://click.palletsprojects.com/en/7.x/), which supports autocompletion.\nThe details of enabling shell completion with click vary by shell ([instructions here](https://click.palletsprojects.com/en/7.x/bashcomplete/)), but here is an example that updates the completion in the background.\n\n```bash\nAWS_SSO_UTIL_COMPLETE_SCRIPT_DIR=~/.local/share/aws-sso-util\nAWS_SSO_UTIL_COMPLETE_SCRIPT=$AWS_SSO_UTIL_COMPLETE_SCRIPT_DIR/complete.sh\nif which aws-sso-util > /dev/null; then\n  mkdir -p $AWS_SSO_UTIL_COMPLETE_SCRIPT_DIR\n  (_AWS_SSO_UTIL_COMPLETE=source_bash aws-sso-util > $AWS_SSO_UTIL_COMPLETE_SCRIPT.tmp ;\n    mv $AWS_SSO_UTIL_COMPLETE_SCRIPT.tmp $AWS_SSO_UTIL_COMPLETE_SCRIPT &)\n  if [ -f $AWS_SSO_UTIL_COMPLETE_SCRIPT ]; then\n    source $AWS_SSO_UTIL_COMPLETE_SCRIPT\n  fi\nfi\n```\n\n## Configuring `.aws/config`\n\nRead the full docs for `aws-sso-util configure` [here](docs/configure.md).\n\n`aws-sso-util configure` has two subcommands, `aws-sso-util configure profile` for configuring a single profile, and `aws-sso-util configure populate` to add _all_ your permissions as profiles, in whatever region(s) you want (with highly configurable profile names).\n\nYou probably want to set the environment variables `AWS_DEFAULT_SSO_START_URL` and `AWS_DEFAULT_SSO_REGION`, which will inform these commands of your start url and SSO region (that is, the region that you've configured AWS SSO in), so that you don't have to pass them in as parameters every time.\n\n`aws-sso-util configure profile` takes a profile name and prompts you with the accounts and roles you have access to, to configure that profile.\n\n`aws-sso-util configure populate` takes one or more regions, and generates a profile for each account+role+region combination.\nThe profile names are completely customizable.\n\n## Logging in and out\n\nRead the full docs for `aws-sso-util login` and `aws-sso-util logout` [here](docs/login.md).\n\nA problem with [`aws sso login`](https://awscli.amazonaws.com/v2/documentation/api/latest/reference/sso/login.html) is that it's required to operate on a profile, that is, you have to tell it to log in to AWS SSO *plus some account and role.*\nBut the whole point of AWS SSO is that you log in once for *many* accounts and roles.\nYou could have a particular account and role set up in your default profile, but I prefer not to have a default profile so that I'm always explicitly selecting a profile and never accidentally end up in the default by mistake.\n`aws-sso-util login` solves this problem by letting you *just log in* without having to think about where you'll be using those credentials.\n\n## Adding AWS SSO support to AWS SDKs\n\nRead the full docs for `aws-sso-util credential-process` [here](docs/credential-process.md).\n\nNot all AWS SDKs have support for AWS SSO (which will change eventually).\nHowever, they all have support for `credential_process`, which allows an external process to provide credentials.\n`aws-sso-util credential-process` uses this to allow these SDKs to get credentials from AWS SSO.\nIt's added automatically (by default) by the `aws-sso-util configure` commands.\n\n## Looking up identifiers and assignments\n\nRead the full docs for `aws-sso-util lookup` and `aws-sso-util assignments` [here](docs/lookup.md).\n\nWhen you're creating assignments through the API or CloudFormation, you're required to use identifiers like the instance ARN, the principal ID, etc.\nThese identifiers aren't readily available through the console, and the principal IDs are not the IDs you're familiar with.\n`aws-sso-util lookup` allows you to get these identifers, even en masse.\n\nThere is no simple API for retrieving all assignments or even a decent subset.\nThe current best you can do is [list all the users with a particular PermissionSet on a particular account](https://docs.aws.amazon.com/singlesignon/latest/APIReference/API_ListAccountAssignments.html).\n`aws-sso-util assigments` takes the effort out of looping over the necessary APIs.\n\n## CloudFormation support\n\n> :warning: The CloudFormation functionality is experimental. I have found the `AWS::SSO::Assignment` resources to be a little brittle and I am not yet confident they will work well if a particular assignment gets moved between nested stacks. This warning will be removed once I have confidence in the functionality.\n\nYou'll want to read the full docs [here](docs/cloudformation.md).\n\nAWS SSO's CloudFormation support currently only includes [`AWS::SSO::Assignment`](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-sso-assignment.html), which means for every combination of principal (group or user), permission set, and target (AWS account), you need a separate CloudFormation resource.\nAdditionally, AWS SSO does not support OUs as targets, so you need to specify every account separately.\n\nObviously, this gets verbose, and even an organization of moderate size is likely to have tens of thousands of assignments.\n`aws-sso-util cfn` provides two mechanisms to make this concise.\n\nI look forward to discarding this part of the tool once there are two prerequisites:\n1. OUs as targets for assignments\n2. An `AWS::SSO::AssignmentGroup` resource that allows specifications of multiple principals, permission sets, and targets, and performs the combinatorics directly.\n\n### CloudFormation Macro\n`aws-sso-util` defines a resource format for an AssignmentGroup that is a combination of multiple principals, permission sets, and targets, and provides a CloudFormation Macro you can deploy that lets you use this resource in your templates.\n\n### Client-side generation\n\nI am against client-side generation of CloudFormation templates, but if you don't want to trust this 3rd party macro, you can generate the CloudFormation templates directly.\n\n`aws-sso-util cfn` takes one or more input files, and for each input file, generates a CloudFormation template and potentially one or more child templates.\nThese templates can then be packaged and uploaded using [`aws cloudformation package`](https://awscli.amazonaws.com/v2/documentation/api/latest/reference/cloudformation/package.html) or [the SAM CLI](https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-install.html), for example.\n\nThe input files can either be templates using the Macro (using the `--macro` flag), or somewhat simpler configuration files using a different syntax.\nThese configuration files can define permission sets inline, have references that turn into template parameters, and you can provide a base template that the resulting resources are layered on top of.\n",
    'author': 'Ben Kehoe',
    'author_email': 'ben@kehoe.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/benkehoe/aws-sso-util',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
from collections import namedtuple, OrderedDict
from pathlib import Path
import logging
import sys
import os

import boto3

import click

from .. import api_utils
from ..cfn.config import Config, ConfigError, validate_config
from ..cfn import resources, templates, macro, utils

from .utils import configure_logging

Input = namedtuple("Input", ["base_path", "stem", "assignments", "permission_sets", "base_template"])

TemplateProcessInput = namedtuple("TemplateProcessInput", [
    "base_path",
    "base_stem",
    "base_template",
    "max_stack_resources",
    "items"
])
TemplateProcessInputItem = namedtuple("TemplateProcessInput", [
    "stem",
    "resource_collection",
])

LOGGER = logging.getLogger(__name__)

def param_loader(ctx, param, value):
    if not value:
        return {}
    d = []
    for p in value.split(","):
        if "=" in p:
            d.append((p.split("=", 1)))
        else:
            d.append((p, None))
    return dict(d)

@click.command()
@click.argument("config_file", required=True, nargs=-1, type=click.File("r"))

@click.option("--macro", is_flag=True, help="Process templates with macro")

@click.option("--profile", help="AWS profile to use to retrieve SSO instance and/or accounts from OUs")
@click.option("--sso-instance", "--ins", help="If not provided, will be retrieved from your account")

@click.option("--template-file-suffix")
@click.option("--output-dir")

@click.option("--base-template-file", type=click.File("r"), help="Base template to build from")
@click.option("--template-parameters", callback=param_loader, help="String-type parameters on the template")

@click.option("--max-resources-per-template", type=int, default=templates.MAX_RESOURCES_PER_TEMPLATE)
@click.option("--max-concurrent-assignments", type=int, default=templates.MAX_CONCURRENT_ASSIGNMENTS)

@click.option("--verbose", "-v", count=True)
def generate_template(
        config_file,
        macro,
        profile,
        sso_instance,
        template_file_suffix,
        output_dir,
        base_template_file,
        template_parameters,
        max_resources_per_template,
        max_concurrent_assignments,
        verbose):
    configure_logging(LOGGER, verbose)


    if macro and base_template_file:
        raise click.UsageError("--base-template-file not allowed with --macro")
    if macro and template_parameters:
        raise click.UsageError("--template-parameters not allowed with --macro")

    session = []
    def session_fetcher():
        if not session:
            session.append(boto3.Session(profile_name=profile))
        return session[0]

    ids = api_utils.Ids(session_fetcher, sso_instance, identity_store_id=None)

    ou_fetcher = lambda ou, recursive: api_utils.get_accounts_for_ou(session, ou, recursive)

    if not template_file_suffix:
        template_file_suffix = ".yaml"
    elif not template_file_suffix.endswith(".yaml"):
        template_file_suffix = template_file_suffix + ".yaml"

    if base_template_file:
        base_template = utils.load_yaml(base_template_file)
        base_template_path = Path(base_template_file.name).resolve()
        prev_len = len(config_file)
        config_file = [c for c in config_file if Path(c.name).resolve() != base_template_path]
        if len(config_file) != prev_len:
            LOGGER.debug("Removed base template file from list of config files")
    else:
        base_template = None

    if macro:
        template_process_inputs = process_macro(
            config_file=config_file,
            session=session,
            ids=ids,
            ou_fetcher=ou_fetcher,
            template_file_suffix=template_file_suffix,
            output_dir=output_dir,
            max_resources_per_template=max_resources_per_template,
            max_concurrent_assignments=max_concurrent_assignments,
        )
    else:
        template_process_inputs = process_config(
            config_file=config_file,
            session=session,
            ids=ids,
            ou_fetcher=ou_fetcher,
            template_file_suffix=template_file_suffix,
            output_dir=output_dir,
            base_template=base_template,
            max_resources_per_template=max_resources_per_template,
            max_concurrent_assignments=max_concurrent_assignments,
        )

    templates_to_write = process_templates(
        template_process_inputs=template_process_inputs,
        template_file_suffix=template_file_suffix,
        max_resources_per_template=max_resources_per_template,
        max_concurrent_assignments=max_concurrent_assignments
    )

    write_templates(templates_to_write)

def process_config(
    config_file,
    session,
    ids,
    ou_fetcher,
    template_file_suffix,
    output_dir,
    base_template,
    max_resources_per_template,
    max_concurrent_assignments):
    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading config file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        data = utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Config file contents:\n{utils.dump_yaml(data)}")

        config = Config()
        config.load(data)

        try:
            validate_config(config, ids)
        except ConfigError as e:
            LOGGER.fatal(f"{e!s} in {config_file_path}")
            sys.exit(1)

        resource_collection = resources.get_resources_from_config(
            config,
            ou_fetcher=ou_fetcher)

        LOGGER.info(f"Generated {len(resource_collection.assignments)} assignments")

        max_stack_resources = templates.get_max_number_of_child_stacks(
            resource_collection.num_resources,
            max_resources_per_template=max_resources_per_template)

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=stem,
                resource_collection=resource_collection
            )]
        )
    return template_process_inputs

def process_macro(
        config_file,
        session,
        ids,
        ou_fetcher,
        template_file_suffix,
        output_dir,
        max_resources_per_template,
        max_concurrent_assignments):

    template_process_inputs = {}

    for config_file_fp in config_file:
        LOGGER.info(f"Loading template file {config_file_fp.name}")
        config_file_path = Path(config_file_fp.name)
        if output_dir:
            base_path = Path(output_dir)
        else:
            base_path = config_file_path.parent / "templates"
        stem = config_file_path.stem

        input_template = utils.load_yaml(config_file_fp)
        LOGGER.debug(f"Input template:\n{utils.dump_yaml(input_template)}")

        LOGGER.info("Extracting resources from template")
        base_template, max_stack_resources, resource_collection_dict = macro.process_template(input_template,
                session=session,
                ids=ids,
                max_resources_per_template=max_resources_per_template)

        template_process_inputs[config_file_path] = TemplateProcessInput(
            base_path=base_path,
            base_stem=stem,
            base_template=base_template,
            max_stack_resources=max_stack_resources,
            items=[TemplateProcessInputItem(
                stem=resource_name,
                resource_collection=resource_collection
            ) for resource_name, resource_collection in resource_collection_dict.items()]
        )

    return template_process_inputs

def process_templates(
        template_process_inputs,
        template_file_suffix,
        max_resources_per_template,
        max_concurrent_assignments):
    templates_to_write = {}

    for name, template_process_input in template_process_inputs.items():
        LOGGER.info(f"Generating templates for {name}")
        parent_template_to_write = template_process_input.base_template or {}
        all_children = []
        for template_process_input_item in template_process_input.items:
            num_parent_resources = len(parent_template_to_write.get("Resources", {})) + template_process_input.max_stack_resources

            parent_template = templates.resolve_templates(
                template_process_input_item.resource_collection.assignments,
                template_process_input_item.resource_collection.permission_sets,
                max_resources_per_template=max_resources_per_template,
                num_parent_resources=num_parent_resources,
            )

            template_collection = parent_template.get_templates(
                template_process_input.base_path,
                ".",
                template_process_input_item.stem,
                template_file_suffix,
                base_template=parent_template_to_write,
                parameters=None,
                max_concurrent_assignments=max_concurrent_assignments,
                path_joiner=os.path.join

            )
            parent_template_to_write = template_collection.parent.template
            LOGGER.debug(f"Intermediate parent template\n{utils.dump_yaml(parent_template_to_write)}")

            all_children.extend(template_collection.children)

        templates_to_write[name] = templates.TemplateCollection(
            parent=templates.WritableTemplate(
                path=os.path.join(
                    template_process_input.base_path,
                    f"{template_process_input.base_stem}{template_file_suffix}"),
                template=parent_template_to_write,
            ),
            children=all_children
        )
    return templates_to_write

def write_templates(templates_to_write):
    for name, template_collection_to_write in templates_to_write.items():
        parent_path = template_collection_to_write.parent.path
        parent_data = template_collection_to_write.parent.template

        for child_path, child_data in template_collection_to_write.children:
            LOGGER.info(f"Writing child template at path {child_path}")
            Path(child_path).parent.mkdir(parents=True, exist_ok=True)
            with open(child_path, "w") as fp:
                utils.dump_yaml(child_data, fp)

        LOGGER.info(f"Writing template for {name} at path {parent_path}")
        Path(parent_path).parent.mkdir(parents=True, exist_ok=True)
        with open(parent_path, "w") as fp:
            utils.dump_yaml(parent_data, fp)


if __name__ == "__main__":
    generate_template(prog_name="python -m aws_sso_util.cli.cfn")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This code is based on the code for the AWS CLI v2's `aws sso login` functionality
# https://github.com/aws/aws-cli/tree/v2/awscli/customizations/sso

import os
import sys
import webbrowser
import logging
import datetime
import uuid

import boto3
import botocore
from botocore.credentials import JSONFileCache

from .vendored_botocore.utils import SSOTokenFetcher
from .vendored_botocore.credentials import SSOCredentialFetcher

from .exceptions import InvalidSSOConfigError, AuthDispatchError, AuthenticationNeededError
from .browser import OpenBrowserHandler, non_interactive_auth_raiser

SSO_TOKEN_DIR = os.path.expanduser(
    os.path.join('~', '.aws', 'sso', 'cache')
)

CREDENTIALS_CACHE_DIR = os.path.expanduser(
    os.path.join('~', '.aws', 'cli', 'cache')
)

LOGGER = logging.getLogger(__name__)

def get_token_fetcher(session, sso_region, interactive=False, token_cache=None,
                     on_pending_authorization=None):
    if hasattr(session, '_session'): #boto3 Session
        session = session._session

    if token_cache is None:
        token_cache = JSONFileCache(SSO_TOKEN_DIR)

    if on_pending_authorization is None:
        if interactive:
            on_pending_authorization = OpenBrowserHandler(
                outfile=sys.stderr,
                open_browser=webbrowser.open_new_tab,
            )
        else:
            on_pending_authorization = non_interactive_auth_raiser

    token_fetcher = SSOTokenFetcher(
        sso_region=sso_region,
        client_creator=session.create_client,
        cache=token_cache,
        on_pending_authorization=on_pending_authorization,
    )
    return token_fetcher

def get_token_loader(session, sso_region, interactive=False, token_cache=None,
                     on_pending_authorization=None, force_refresh=False):
    token_fetcher = get_token_fetcher(
        session=session,
        sso_region=sso_region,
        interactive=interactive,
        token_cache=token_cache,
        on_pending_authorization=on_pending_authorization,
    )
    return get_token_loader_from_token_fetcher(token_fetcher, force_refresh=force_refresh)

def get_token_loader_from_token_fetcher(token_fetcher, force_refresh=False):
    def token_loader(start_url):
        token_response = token_fetcher.fetch_token(
            start_url=start_url,
            force_refresh=force_refresh
        )
        LOGGER.debug('TOKEN RESPONSE: {}'.format(token_response))
        return token_response['accessToken']

    return token_loader


def get_credentials(session, start_url, sso_region, account_id, role_name, token_loader=None, cache=None):
    if hasattr(session, '_session'): #boto3 Session
        session = session._session
    if not token_loader:
        token_loader = get_token_loader(session, sso_region)

    if cache is None:
        cache = JSONFileCache(CREDENTIALS_CACHE_DIR)

    credential_fetcher = SSOCredentialFetcher(
        start_url=start_url,
        sso_region=sso_region,
        role_name=role_name,
        account_id=account_id,
        client_creator=session.create_client,
        cache=cache,
        token_loader=token_loader,
    )

    return credential_fetcher.fetch_credentials()

def get_botocore_session(start_url, sso_region, account_id, role_name):
    botocore_session = botocore.session.Session()

    profile_name = str(uuid.uuid4())
    load_config = lambda: {
        "profiles": {
            profile_name: {
                "sso_start_url": start_url,
                "sso_region": sso_region,
                "sso_account_id": account_id,
                "sso_role_name": role_name,
            }
        }
    }
    sso_provider = botocore.credentials.SSOProvider(
        load_config=load_config,
        client_creator=botocore_session.create_client,
        profile_name=profile_name
    )

    botocore_session.register_component(
        'credential_provider',
        botocore.credentials.CredentialResolver([sso_provider])
    )

    return botocore_session

def get_boto3_session(start_url, sso_region, account_id, role_name, region):
    botocore_session = get_botocore_session(start_url, sso_region, account_id, role_name)

    session = boto3.Session(botocore_session=botocore_session, region_name=region)

    return session



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys
import os
import argparse
import logging
import json
import subprocess
import shlex
import re
from collections import namedtuple

import botocore
from botocore.session import Session
from botocore.exceptions import ClientError, ProfileNotFound

import click

from ..sso import get_token_loader
from ..config import find_instances, SSOInstance
from ..vendored_botocore.config_file_writer import ConfigFileWriter, write_values, get_config_filename
from .utils import configure_logging

from .configure_profile import (
    DEFAULT_START_URL_VARS,
    DEFAULT_SSO_REGION_VARS,
    DEFAULT_REGION_VARS,
    DISABLE_CREDENTIAL_PROCESS_VAR,
    CREDENTIAL_PROCESS_NAME_VAR,
    SET_CREDENTIAL_PROCESS_DEFAULT
)

DEFAULT_SEPARATOR = "."

LOGGER = logging.getLogger(__name__)

ConfigParams = namedtuple("ConfigParams", ["profile_name", "account_name", "account_id", "role_name", "region"])

def get_short_region(region):
    area, direction, num = region.split("-")
    dir_abbr = {
        "north": "no",
        "northeast": "ne",
        "east": "ea",
        "southeast": "se",
        "south": "so",
        "southwest": "sw",
        "west": "we",
        "northwest": "nw",
        "central": "ce",
    }
    return "".join([area, dir_abbr.get(direction, direction), num])

KNOWN_COMPONENTS = [
    "account_name",
    "account_id",
    "account_number",
    "role_name",
    "region",
    "short_region",
]
def generate_profile_name_format(input, separator, region_style):
    def process_component(c):
        if c == "default_style_region":
            if region_style == "short":
                c = "short_region"
            else:
                c = "region"
        if c in KNOWN_COMPONENTS:
            return "{" + c + "}"
        else:
            return c
    region_format = separator.join(process_component(c) for c in input.split(","))
    no_region_format = separator.join(process_component(c) for c in input.split(",") if c not in ["default_style_region", "region", "short_region"])
    return region_format, no_region_format

def get_formatter(include_region, region_format, no_region_format):
    def proc_kwargs(kwargs):
        kwargs["short_region"] = get_short_region(kwargs["region"])
        kwargs["account_number"] = kwargs["account_id"]
        return kwargs
    if include_region == "default":
        def formatter(i, **kwargs):
            kwargs = proc_kwargs(kwargs)
            if i == 0:
                return no_region_format.format(**kwargs)
            else:
                return region_format.format(**kwargs)
        return formatter
    elif include_region == "always":
        def formatter(i, **kwargs):
            kwargs = proc_kwargs(kwargs)
            return region_format.format(**kwargs)
        return formatter
    else:
        raise ValueError("Unknown include_region value {}".format(include_region))

PROCESS_FORMATTER_ARGS = [
    "account_name",
    "account_id",
    "role_name",
    "region",
    "short_region",
    "region_index",
]
def get_process_formatter(command):
    def formatter(i, **kwargs):
        kwargs["region_index"] = str(i)
        kwargs["short_region"] = get_short_region(kwargs["region"])
        run_args = shlex.split(command)
        for component in PROCESS_FORMATTER_ARGS:
            run_args.append(kwargs[component])
        try:
            result = subprocess.run(shlex.join(run_args), shell=True, stdout=subprocess.PIPE, check=True)
        except subprocess.CalledProcessError as e:
            lines = [
                "Profile name process failed ({})".format(e.returncode)
            ]
            if e.stdout:
                lines.append(e.stdout.decode("utf-8"))
            if e.stderr:
                lines.append(e.stderr.decode("utf-8"))
            LOGGER.error("\n".join(lines))
            raise e
        return result.stdout.decode("utf-8").strip()
    return formatter

def get_trim_formatter(account_name_patterns, role_name_patterns, formatter):
    def trim_formatter(i, **kwargs):
        for pattern in account_name_patterns:
            kwargs["account_name"] = re.sub(pattern, "", kwargs["account_name"])
        for pattern in role_name_patterns:
            kwargs["role_name"] = re.sub(pattern, "", kwargs["role_name"])
        return formatter(i, **kwargs)
    return trim_formatter

@click.command()
@click.option("--sso-start-url", "-u")
@click.option("--sso-region")

@click.option("--region", "-r", "regions", multiple=True)

@click.option("--dry-run", is_flag=True, help="Print the config to stdout")

@click.option("--config-default", "-c", multiple=True)
@click.option("--existing-config-action", type=click.Choice(["keep", "overwrite", "discard"]), default="keep")

@click.option("--components", "profile_name_components", default="account_name,role_name,default_style_region")
@click.option("--separator", "--sep", "profile_name_separator", help=f"Default is {DEFAULT_SEPARATOR}")
@click.option("--include-region", "profile_name_include_region", type=click.Choice(["default", "always"]), default="default")
@click.option("--region-style", "profile_name_region_style", type=click.Choice(["short", "long"]), default="short")
@click.option("--trim-account-name", "profile_name_trim_account_name_patterns", multiple=True, default=[], help="Regex to remove from account names")
@click.option("--trim-role-name", "profile_name_trim_role_name_patterns", multiple=True, default=[], help="Regex to remove from account names")
@click.option("--profile-name-process")

@click.option("--credential-process/--no-credential-process", default=None)

@click.option("--force-refresh", is_flag=True, help="Re-login")
@click.option("--verbose", "-v", count=True)
def populate_profiles(
        sso_start_url,
        sso_region,
        regions,
        dry_run,
        config_default,
        existing_config_action,
        profile_name_components,
        profile_name_separator,
        profile_name_include_region,
        profile_name_region_style,
        profile_name_trim_account_name_patterns,
        profile_name_trim_role_name_patterns,
        profile_name_process,
        credential_process,
        force_refresh,
        verbose):

    configure_logging(LOGGER, verbose)

    missing = []

    instances, specifier, all_instances = find_instances(
        profile_name=None,
        profile_source=None,
        start_url=sso_start_url,
        start_url_source="CLI input",
        region=sso_region,
        region_source="CLI input",
        start_url_vars=DEFAULT_START_URL_VARS,
        region_vars=DEFAULT_SSO_REGION_VARS,
    )

    if not instances:
        if all_instances:
            LOGGER.fatal((
                f"No AWS SSO config matched {specifier.to_str(region=True)} " +
                f"from {SSOInstance.to_strs(all_instances)}"))
        else:
            LOGGER.fatal("No AWS SSO config found")
        sys.exit(1)

    if len(instances) > 1:
        LOGGER.fatal(f"Found {len(instances)} SSO configs, please specify one: {SSOInstance.to_strs(instances)}")
        sys.exit(1)

    instance = instances[0]

    if not regions:
        for var_name in DEFAULT_REGION_VARS:
            value = os.environ.get(var_name)
            if value:
                LOGGER.debug(f"Got default region {value} from {var_name}")
                regions = [value]
                break
    if not regions:
        missing.append("--region")

    if missing:
        raise click.UsageError("Missing arguments: {}".format(", ".join(missing)))

    if config_default:
        config_default = dict(v.split("=", 1) for v in config_default)
    else:
        config_default = {}

    if not profile_name_separator:
        profile_name_separator = os.environ.get("AWS_CONFIGURE_SSO_DEFAULT_PROFILE_NAME_SEPARATOR") or DEFAULT_SEPARATOR

    if profile_name_process:
        profile_name_formatter = get_process_formatter(profile_name_process)
    else:
        region_format, no_region_format = generate_profile_name_format(profile_name_components, profile_name_separator, profile_name_region_style)
        LOGGER.debug("Profile name format (region):    {}".format(region_format))
        LOGGER.debug("Profile name format (no region): {}".format(no_region_format))
        profile_name_formatter = get_formatter(profile_name_include_region, region_format, no_region_format)
        if profile_name_trim_account_name_patterns or profile_name_trim_role_name_patterns:
            profile_name_formatter = get_trim_formatter(profile_name_trim_account_name_patterns, profile_name_trim_role_name_patterns, profile_name_formatter)

    try:
        profile_name_formatter(0, account_name="foo", account_id="bar", role_name="baz", region="us-east-1")
    except Exception as e:
        raise click.UsageError("Invalid profile name format: {}".format(e))

    session = Session()

    token_loader = get_token_loader(session,
            instance.region,
            interactive=True,
            force_refresh=force_refresh)

    LOGGER.info(f"Logging in to {instance.start_url}")
    token = token_loader(instance.start_url)

    LOGGER.debug("Token: {}".format(token))

    config = botocore.config.Config(
        region_name=instance.region,
        signature_version=botocore.UNSIGNED,
    )
    client = session.create_client("sso", config=config)

    LOGGER.info("Gathering accounts and roles")
    accounts = []
    list_accounts_args = {
        "accessToken": token
    }
    while True:
        response = client.list_accounts(**list_accounts_args)

        accounts.extend(response["accountList"])

        next_token = response.get("nextToken")
        if not next_token:
            break
        else:
            list_accounts_args["nextToken"] = response["nextToken"]

    LOGGER.debug("Account list: {} {}".format(len(accounts), accounts))

    configs = []
    for account in accounts:
        LOGGER.debug("Getting roles for {}".format(account["accountId"]))
        list_role_args = {
            "accessToken": token,
            "accountId": account["accountId"],
        }

        while True:
            response = client.list_account_roles(**list_role_args)

            for role in response["roleList"]:
                for i, region in enumerate(regions):
                    profile_name = profile_name_formatter(i,
                        account_name=account["accountName"],
                        account_id=account["accountId"],
                        role_name=role["roleName"],
                        region=region,
                    )
                    configs.append(ConfigParams(profile_name, account["accountName"], account["accountId"], role["roleName"], region))

            next_token = response.get("nextToken")
            if not next_token:
                break
            else:
                list_role_args["nextToken"] = response["nextToken"]


    configs.sort(key=lambda v: v.profile_name)

    LOGGER.debug("Got configs: {}".format(configs))

    if not dry_run:
        LOGGER.info("Writing {} profiles to {}".format(len(configs), get_config_filename(session)))

        config_writer = ConfigFileWriter()
        def write_config(profile_name, config_values):
            write_values(session, config.profile_name, config_values, config_file_writer=config_writer)
    else:
        LOGGER.info("Dry run for {} profiles".format(len(configs)))
        def write_config(profile_name, config_values):
            lines = [
                "[profile {}]".format(profile_name)
            ]
            for key, value in config_values.items():
                lines.append("{} = {}".format(key, value))
            lines.append("")
            print("\n".join(lines))

    for config in configs:
        LOGGER.debug("Processing config: {}".format(config))

        config_values = {}
        existing_profile = False
        existing_config = {}
        if existing_config_action != "discard":
            try:
                existing_config = Session(profile=config.profile_name).get_scoped_config()
                config_values.update(existing_config)
                existing_profile = True
            except ProfileNotFound:
                pass

        config_values.update({
            "sso_start_url": instance.start_url,
            "sso_region": instance.region,
            "sso_account_name": config.account_name,
            "sso_account_id": config.account_id,
            "sso_role_name": config.role_name,
            "region": config.region,
        })


        for k, v in config_default.items():
            if k in existing_config and existing_config_action in ["keep"]:
                continue
            config_values[k] = v

        if credential_process is not None:
            set_credential_process = credential_process
        elif os.environ.get(DISABLE_CREDENTIAL_PROCESS_VAR, "").lower() in ["1", "true"]:
            set_credential_process = False
        else:
            set_credential_process = SET_CREDENTIAL_PROCESS_DEFAULT

        if set_credential_process:
            credential_process_name = os.environ.get(CREDENTIAL_PROCESS_NAME_VAR) or "aws-sso-util credential-process"
            config_values["credential_process"] = f"{credential_process_name} --profile {config.profile_name}"
        elif set_credential_process is False:
            config_values.pop("credential_process", None)

        LOGGER.debug("Config values for profile {}: {}".format(config.profile_name, config_values))

        write_config(config.profile_name, config_values)


if __name__ == "__main__":
    populate_profiles(prog_name="python -m aws_sso_util.cli.populate_profiles")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import uuid
import random
import string

"""
instance arn: arn:aws:sso:::instance/ssoins-{16 hex}
identity store: d-{10 hex}
start url: https://d-X.awsapps.com/start
group: uuid
user: uuid
permission set: arn:aws:sso:::permissionSet/ssoins-X/ps-{16 hex}
ou: ou- 4 alphanum - 8 alphanum
account: 123456789012
"""

hex = '0123456789abcdef'
alphanum = string.ascii_lowercase + string.digits

def sample(lst, length):
    return ''.join(random.choice(lst) for _ in range(length))

instance_id = 'ssoins-{}'.format(sample(hex, 16))
identity_store = 'd-{}'.format(sample(hex, 10))

print('instance:       arn:aws:sso:::instance/{}'.format(instance_id))

print('identity store: {}'.format(identity_store))

print('start url:      https://{}.awsapps.com/start'.format(identity_store))

print('user/group:     {}'.format(uuid.uuid4()))

print('permission-set: arn:aws:sso:::permissionSet/{}/ps-{}'.format(instance_id, sample(hex, 16)))

print('OU:             ou-{}-{}'.format(sample(alphanum, 4), sample(alphanum, 8)))

print('account:        {}'.format(sample(string.digits, 12)))





# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import subprocess
import sys
import logging
import textwrap

import botocore
from botocore.exceptions import ProfileNotFound

import click

from ..config import find_instances, SSOInstance
from ..vendored_botocore.config_file_writer import write_values
from .utils import configure_logging

LOGGER = logging.getLogger(__name__)

DEFAULT_START_URL_VARS  = ["AWS_CONFIGURE_SSO_DEFAULT_SSO_START_URL", "AWS_CONFIGURE_DEFAULT_SSO_START_URL"]
DEFAULT_SSO_REGION_VARS = ["AWS_CONFIGURE_SSO_DEFAULT_SSO_REGION",    "AWS_CONFIGURE_DEFAULT_SSO_REGION"]
DEFAULT_REGION_VARS = ["AWS_CONFIGURE_DEFAULT_REGION", "AWS_DEFAULT_REGION"]
DISABLE_CREDENTIAL_PROCESS_VAR = "AWS_CONFIGURE_SSO_DISABLE_CREDENTIAL_PROCESS"

CREDENTIAL_PROCESS_NAME_VAR= "AWS_SSO_CREDENTIAL_PROCESS_NAME"

SET_CREDENTIAL_PROCESS_DEFAULT = True

@click.command("configure-profile")
@click.argument("profile")
@click.option("--sso-start-url", "-u")
@click.option("--sso-region")
@click.option("--account-id", "-a", "account")
@click.option("--role-name", "-r", "role")
@click.option("--region")
@click.option("--output", "-o")
@click.option("--config-default", "-c", multiple=True)
@click.option("--existing-config-action", type=click.Choice(["keep", "overwrite", "discard"]), default="keep")
@click.option("--interactive/--non-interactive", default=True)
@click.option("--credential-process/--no-credential-process", default=None)
@click.option("--verbose", "-v", count=True)
def configure_profile(
        profile,
        sso_start_url,
        sso_region,
        account,
        role,
        region,
        output,
        config_default,
        existing_config_action,
        interactive,
        credential_process,
        verbose):
    configure_logging(LOGGER, verbose)

    instances, specifier, all_instances = find_instances(
        profile_name=None,
        profile_source=None,
        start_url=sso_start_url,
        start_url_source="CLI input",
        region=sso_region,
        region_source="CLI input",
        start_url_vars=DEFAULT_START_URL_VARS,
        region_vars=DEFAULT_SSO_REGION_VARS,
    )

    if not instances and all_instances:
        LOGGER.fatal((
            f"No AWS SSO config matched {specifier.to_str(region=True)} " +
            f"from {SSOInstance.to_strs(all_instances)}"))
        sys.exit(1)


    if len(instances) > 1:
        LOGGER.fatal(f"Found {len(instances)} SSO configs, please specify one or use --all: {SSOInstance.to_strs(instances)}")
        sys.exit(1)

    instance = instances[0] if instances else None

    sso_start_url = instance.start_url if instance else None
    sso_region    = instance.region    if instance else None

    if config_default:
        config_default = dict(v.split("=", 1) for v in config_default)
    else:
        config_default = {}

    session = botocore.session.Session(profile=profile)

    config_values = {}
    existing_profile = False
    existing_config = {}
    if existing_config_action != "discard":
        try:
            existing_config = session.get_scoped_config()
            config_values.update(existing_config)
            existing_profile = True
        except ProfileNotFound:
            pass

    if sso_start_url:
        config_values["sso_start_url"] = sso_start_url

    if sso_region:
        config_values["sso_region"] = sso_region

    if account:
        config_values["sso_account_id"] = account

    if role:
        config_values["sso_role_name"] = role

    if region:
        config_values["region"] = region
    elif "region" not in config_values:
        for var_name in DEFAULT_REGION_VARS:
            value = os.environ.get(var_name)
            if value:
                LOGGER.debug(f"Got default region {value} from {var_name}")
                config_values["region"] = value
                break

    if output:
        config_values["output"] = output

    for k, v in config_default.items():
        if k in existing_config and existing_config_action in ["keep"]:
            continue
        config_values[k] = v

    if credential_process is not None:
        set_credential_process = credential_process
    elif os.environ.get(DISABLE_CREDENTIAL_PROCESS_VAR, "").lower() in ["1", "true"]:
        set_credential_process = False
    else:
        set_credential_process = SET_CREDENTIAL_PROCESS_DEFAULT

    if set_credential_process:
        credential_process_name = os.environ.get(CREDENTIAL_PROCESS_NAME_VAR) or "aws-sso-util credential-process"
        config_values["credential_process"] = f"{credential_process_name} --profile {profile}"
    elif set_credential_process is False:
        config_values.pop("credential_process", None)

    required_keys = [
        "sso_start_url",
        "sso_region",
        "sso_account_id",
        "sso_role_name",
        "region"
    ]

    missing_keys = [k for k in required_keys if k not in config_values]

    if missing_keys and not interactive:
        LOGGER.error(f"Missing profile options {', '.join(missing_keys)}")
        sys.exit(1)
    else:
        LOGGER.debug(f"Missing keys: {', '.join(missing_keys)}")

    write_values(session, profile, config_values)

    if not missing_keys:
        return

    try:
        result = subprocess.run(["aws", "--version"], capture_output=True)
        cli_version = parse_cli_version(result.stdout.decode("utf-8"))
        if cli_version.startswith("1."):
            LOGGER.warn(textwrap.dedent(f"""\
            Your profile has been written, but is not complete.
            You have the AWS CLI version {cli_version}, which does not support AWS SSO.
            If you install the AWS CLI v2, aws-sso-util configure profile can interactively
            prompt you for the necessary fields. Details here:
            https://docs.aws.amazon.com/cli/latest/userguide/install-cliv2.html"""))
            sys.exit(2)
    except FileNotFoundError:
        LOGGER.warn(textwrap.dedent("""\
            Your profile has been written, but is not complete.
            If you install the AWS CLI v2, aws-sso-util configure profile can interactively
            prompt you for the necessary fields. Details here:
            https://docs.aws.amazon.com/cli/latest/userguide/install-cliv2.html"""))
        sys.exit(2)

    result = subprocess.run(f"aws configure sso --profile {profile}", shell=True)

    if result.returncode:
        # this doesn't appear to work
        # LOGGER.error(f"Interactive configuration existed without success, restoring previous values")
        # write_values(session, profile, existing_config)
        sys.exit(10+result.returncode)

def parse_cli_version(output):
    d = dict(part.split("/", 1) for part in output.split(" "))
    LOGGER.debug(f"AWS CLI version info: {d}")
    return d.get("aws-cli", "UNKNOWN")

if __name__ == "__main__":
    configure_profile(prog_name="python -m aws_sso_util.cli.configure_profile")  #pylint: disable=unexpected-keyword-arg,no-value-for-parameter



# Copyright 2020 Ben Kehoe
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import sys
import textwrap
import webbrowser

from .exceptions import AuthenticationNeededError, AuthDispatchError

class OpenBrowserHandler(object):
    def __init__(self, outfile=None, open_browser=None):
        self._outfile = outfile or sys.stderr
        if open_browser is None:
            open_browser = webbrowser.open_new_tab
        self._open_browser = open_browser

    def __call__(self, userCode, verificationUri,
                 verificationUriComplete, **kwargs):
        message = textwrap.dedent("""\
        AWS SSO login required.
        Attempting to open the SSO authorization page in your default browser.
        If the browser does not open or you wish to use a different device to
        authorize this request, open the following URL:

        {verificationUri}

        Then enter the code:

        {userCode}
        """.format(verificationUri=verificationUri, userCode=userCode))

        print(message, file=sys.stderr)

        disable_browser = os.environ.get('AWS_SSO_DISABLE_BROWSER', '').lower() in ['1', 'true']
        if self._open_browser and not disable_browser:
            try:
                return self._open_browser(verificationUriComplete)
            except AuthenticationNeededError:
                raise
            except Exception as e:
                raise AuthDispatchError('Failed to open browser') from e
                # LOG.debug('Failed to open browser:', exc_info=True)

def non_interactive_auth_raiser(*args, **kwargs):
    raise AuthenticationNeededError
